# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines various text generation classes and methods
"""

import math
import random

import logging
import numpy as np
import pandas as pd

#: list of hex digits for template generation
_HEX_LOWER = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f']

#: list of upper case hex digits for template generation
_HEX_UPPER = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F']

#: list of non-zero digits for template generation
_DIGITS_NON_ZERO = ['1', '2', '3', '4', '5', '6', '7', '8', '9']

#: list of digits for template generation
_DIGITS_ZERO = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

#: list of uppercase letters for template generation
_LETTERS_UPPER = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
                  'Q', 'R', 'T', 'S', 'U', 'V', 'W', 'X', 'Y', 'Z']

#: list of lowercase letters for template generation
_LETTERS_LOWER = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q',
                  'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']

#: list of all letters uppercase and lowercase
_LETTERS_ALL = _LETTERS_LOWER + _LETTERS_UPPER

#: list of alphanumeric chars in lowercase
_ALNUM_LOWER = _LETTERS_LOWER + _DIGITS_ZERO

#: list of alphanumeric chars in uppercase
_ALNUM_UPPER = _LETTERS_UPPER + _DIGITS_ZERO

""" words for ipsum lorem based text generation"""
_WORDS_LOWER = ['lorem', 'ipsum', 'dolor', 'sit', 'amet', 'consectetur', 'adipiscing', 'elit', 'sed', 'do',
                'eiusmod', 'tempor', 'incididunt', 'ut', 'labore', 'et', 'dolore', 'magna', 'aliqua', 'ut',
                'enim', 'ad', 'minim', 'veniam', 'quis', 'nostrud', 'exercitation', 'ullamco', 'laboris',
                'nisi', 'ut', 'aliquip', 'ex', 'ea', 'commodo', 'consequat', 'duis', 'aute', 'irure', 'dolor',
                'in', 'reprehenderit', 'in', 'voluptate', 'velit', 'esse', 'cillum', 'dolore', 'eu', 'fugiat',
                'nulla', 'pariatur', 'excepteur', 'sint', 'occaecat', 'cupidatat', 'non', 'proident', 'sunt',
                'in', 'culpa', 'qui', 'officia', 'deserunt', 'mollit', 'anim', 'id', 'est', 'laborum']

_WORDS_UPPER = ['LOREM', 'IPSUM', 'DOLOR', 'SIT', 'AMET', 'CONSECTETUR', 'ADIPISCING', 'ELIT', 'SED', 'DO',
                'EIUSMOD', 'TEMPOR', 'INCIDIDUNT', 'UT', 'LABORE', 'ET', 'DOLORE', 'MAGNA', 'ALIQUA', 'UT',
                'ENIM', 'AD', 'MINIM', 'VENIAM', 'QUIS', 'NOSTRUD', 'EXERCITATION', 'ULLAMCO', 'LABORIS',
                'NISI', 'UT', 'ALIQUIP', 'EX', 'EA', 'COMMODO', 'CONSEQUAT', 'DUIS', 'AUTE', 'IRURE',
                'DOLOR', 'IN', 'REPREHENDERIT', 'IN', 'VOLUPTATE', 'VELIT', 'ESSE', 'CILLUM', 'DOLORE',
                'EU', 'FUGIAT', 'NULLA', 'PARIATUR', 'EXCEPTEUR', 'SINT', 'OCCAECAT', 'CUPIDATAT', 'NON',
                'PROIDENT', 'SUNT', 'IN', 'CULPA', 'QUI', 'OFFICIA', 'DESERUNT', 'MOLLIT', 'ANIM', 'ID', 'EST',
                'LABORUM']


class TextGenerator(object):
    """ Base class for text generation classes

    """

    def __init__(self):
        self._randomSeed = 42
        self._rngInstance = None

    def __repr__(self):
        return f"TextGenerator(randomSeed={self._randomSeed})"

    def __str__(self):
        return f"TextGenerator(randomSeed={self._randomSeed})"

    def __eq__(self, other):
        return type(self) == type(other) and self._randomSeed == other._randomSeed

    def withRandomSeed(self, seed):
        """ Set the random seed for the text generator

        :param seed: seed value to set
        :return: self
        """
        assert seed is None or type(seed) is int, "expecting an integer seed for Text Generator"
        self._randomSeed = seed
        return self

    @property
    def randomSeed(self):
        """ Get random seed for text generator"""
        return self._randomSeed

    def getNPRandomGenerator(self, forceNewInstance=False):
        """ Get numpy random number generator

        :return: returns random number generator initialized from previously supplied random seed
        """
        assert self._randomSeed is None or type(self._randomSeed) in [int, np.int32, np.int64], \
            f"`random_seed` must be int or int-like not {type(self._randomSeed)}"

        if self._rngInstance is not None and not forceNewInstance:
            return self._rngInstance

        from numpy.random import default_rng
        if self._randomSeed is not None and self._randomSeed not in (-1, -1.0):
            rng = default_rng(seed=self._randomSeed)
        else:
            rng = default_rng()

        if not forceNewInstance:
            self._rngInstance = rng
        return rng

    @staticmethod
    def compactNumpyTypeForValues(listValues):
        """ determine smallest numpy type to represent values

        :param listValues: list or np.ndarray of values to get np.dtype for
        :return: np.dtype that is most compact representation for values provided
        """
        if type(listValues) is list:
            max_value_represented = np.max(np.array(listValues).flatten())
        else:
            max_value_represented = np.max(listValues.flatten()) + 1
        bits_required = math.ceil(math.log2(max_value_represented))

        if bits_required <= 8:
            # for small values, use byte representation
            retval = np.dtype('B')
        else:
            # compute bytes required and raise to nearest power of 2
            bytesRequired = int(math.ceil(bits_required / 8.0))
            retval = np.dtype(f"u{bytesRequired}")
        return retval

    @staticmethod
    def getAsTupleOrElse(v, defaultValue, valueName):
        """ get value v as tuple or return default value

            :param v: value to test
            :param defaultValue: value to use as a default if value of `v` is None. Must be a tuple.
            :param valueName: name of value for debugging and logging purposes
            :returns: return `v` as tuple if not `None` or value of `default_v` if `v` is `None`. If `v` is a single
                      value, returns the tuple (`v`, `v`)"""
        assert v is None or type(v) is int or type(v) is tuple, f"param {valueName} must be an int, a tuple or None"
        assert type(defaultValue) is tuple and len(defaultValue) == 2, "default value must be tuple"

        if type(v) is int:
            return v, v
        elif type(v) is tuple:
            assert len(v) == 2, "expecting tuple of length 2"
            assert type(v[0]) is int and type(v[1]) is int, "expecting tuple with both elements as integers"
            return v
        else:
            assert len(defaultValue) == 2, "must have list or iterable with lenght 2"
            assert type(defaultValue[0]) is int and type(defaultValue[1]) is int, "all elements must be integers"

        return defaultValue


class TemplateGenerator(TextGenerator):  # lgtm [py/missing-equals]
    """This class handles the generation of text from templates

    :param template: template string to use in text generation
    :param escapeSpecialChars: By default special chars in the template have special meaning if unescaped
                               If set to true, then the special meaning requires escape char ``\\``
    :param extendedWordList: if provided, use specified word list instead of default word list

    The template generator generates text from a template to allow for generation of synthetic account card numbers,
    VINs, IBANs and many other structured codes.

    The base value is passed to the template generation and may be used in the generated text. The base value is the
    value the column would have if the template generation had not been applied.

    It uses the following special chars:

    ==========  ======================================
    Chars       Meaning
    ==========  ======================================
    ``\\``       Apply escape to next char.
    v0,v1,..v9  Use base value as an array of values and substitute the `nth` element ( 0 .. 9). Always escaped.
    x           Insert a random lowercase hex digit
    X           Insert an uppercase random hex digit
    d           Insert a random lowercase decimal digit
    D           Insert an uppercase random decimal digit
    a           Insert a random lowercase alphabetical character
    A           Insert a random uppercase alphabetical character
    k           Insert a random lowercase alphanumeric character
    K           Insert a random uppercase alphanumeric character
    n           Insert a random number between 0 .. 255 inclusive. This option must always be escaped
    N           Insert a random number between 0 .. 65535 inclusive. This option must always be escaped
    w           Insert a random lowercase word from the ipsum lorem word set. Always escaped
    W           Insert a random uppercase word from the ipsum lorem word set. Always escaped
    ==========  ======================================

    .. note::
              If escape is used and`escapeSpecialChars` is False, then the following
              char is assumed to have no special meaning.

              If the `escapeSpecialChars` option is set to True, then the following char only has its special
              meaning when preceded by an escape.

              Some options must be always escaped for example  ``\\v``, ``\\n`` and ``\\w``.

              A special case exists for ``\\v`` - if immediately followed by a digit 0 - 9, the underlying base value
              is interpreted as an array of values and the nth element is retrieved where `n` is the digit specified.

    In all other cases, the char itself is used.

    The setting of the `escapeSpecialChars` determines how templates generate data.

    If set to False, then the template ``r"\\dr_\\v"`` will generate the values ``"dr_0"`` ... ``"dr_999"`` when applied
    to the values zero to 999. This conforms to earlier implementations for backwards compatibility.

    If set to True, then the template ``r"dr_\\v"`` will generate the values ``"dr_0"`` ... ``"dr_999"``
    when applied to the values zero to 999. This conforms to the preferred style going forward

    """
    def __init__(self, template, escapeSpecialChars=False, extendedWordList=None):
        assert template is not None, "`template` must be specified"
        super().__init__()

        self._template = template
        self._escapeSpecialMeaning = bool(escapeSpecialChars)
        self._templates = self._splitTemplates(self._template)
        self._wordList = np.array(extendedWordList if extendedWordList is not None else _WORDS_LOWER)
        self._upperWordList = np.array([x.upper() for x in extendedWordList]
                                       if extendedWordList is not None else _WORDS_UPPER)

        self._np_digits_zero = np.array(_DIGITS_ZERO)
        self._np_digits_non_zero = np.array(_DIGITS_NON_ZERO)
        self._np_hex_upper = np.array(_HEX_UPPER)
        self._np_hex_lower = np.array(_HEX_LOWER)
        self._np_alnum_lower = np.array(_ALNUM_LOWER)
        self._np_alnum_upper = np.array(_ALNUM_UPPER)
        self._np_letters_lower = np.array(_LETTERS_LOWER)
        self._np_letters_upper = np.array(_LETTERS_UPPER)
        self._np_letters_all = np.array(_LETTERS_ALL)
        self._lenWords = len(self._wordList)

        # mappings must be mapping from string to tuple(length of mappings, mapping array or list)
        self._templateMappings = {
            'a': (26, self._np_letters_lower),
            'A': (26, self._np_letters_upper),
            'x': (16, self._np_hex_lower),
            'X': (16, self._np_hex_upper),
            'd': (10, self._np_digits_zero),
            'D': (9, self._np_digits_non_zero),
            'k': (36, self._np_alnum_lower),
            'K': (36, self._np_alnum_upper)
        }

        # ensure that each mapping is mapping from string to list or numpy array
        for k, v in self._templateMappings.items():
            assert (k is not None) and isinstance(k, str) and len(k) > 0, "key must be non-empty string"
            assert v is not None and isinstance(v, tuple) and len(v) == 2, "value must be tuple of length 2"
            mapping_length, mappings = v
            assert isinstance(mapping_length, int), "mapping length must be of type int"
            assert isinstance(mappings, (list, np.ndarray)),\
                "mappings are lists or numpy arrays"
            assert mapping_length == 0 or len(mappings) == mapping_length, "mappings must match mapping_length"

        self._templateEscapedMappings = {
            'n': (256, None),
            'N': (65536, None),
            'w': (self._lenWords, self._wordList),
            'W': (self._lenWords, self._upperWordList)
        }

        # ensure that each escaped mapping is mapping from string to None, list or numpy array
        for k, v in self._templateEscapedMappings.items():
            assert (k is not None) and isinstance(k, str) and len(k) > 0, "key must be non-empty string"
            assert v is not None and isinstance(v, tuple) and len(v) == 2, "value must be tuple of length 2"
            mapping_length, mappings = v
            assert isinstance(mapping_length, int), "mapping length must be of type int"
            assert mappings is None or isinstance(mappings, (list, np.ndarray)),\
                "mappings are lists or numpy arrays"

            # for escaped mappings, the mapping can be None in which case the mapping is to the number itself
            # i.e mapping[4] = 4
            assert mappings is None or len(mappings) == mapping_length, "mappings must match mapping_length"

        # get the template metadata - this will be list of metadata entries for each template
        # for each template, metadata will be tuple of number of placeholders followed by list of random bounds
        # to be computed when replacing non static placeholder
        template_info = [self._prepareTemplateStrings(template, escapeSpecialMeaning=escapeSpecialChars)
                                    for template in self._templates]

        self._max_placeholders = max([ x[0] for x in template_info])  # pylint: disable=consider-using-generator
        self._max_rnds_needed = max([ len(x[1]) for x in template_info])  # pylint: disable=consider-using-generator
        self._placeholders_needed = [ x[0] for x in template_info]
        self._template_rnd_bounds = [ x[1] for x in template_info]

    def __repr__(self):
        return f"TemplateGenerator(template='{self._template}')"

    def _splitTemplates(self, templateStr):
        """ Split template string into individual template strings

        :param templateStr: template string
        :return: list of individual template strings


        """
        tmp_template = templateStr.replace(r'\\', '$__escape__').replace(r'\|', '$__sep__')
        results = [x.replace('$__escape__', r'\\').replace('$__sep__', '|') for x in tmp_template.split('|')]
        return results

    @property
    def templates(self):
        """ Get effective templates for text generator"""
        return self._templates

    def _getRandomInt(self, low, high=-1, rng=None):
        """ generate random integer between low and high inclusive

        :param low: low value, if no high value is specified, treat low value as high value and low of 0
        :param high: high value for random number generation
        :param rng: if provided, an instance of a numpy random number generator
        :return: generated value
        """
        if high == -1:
            high = low
            low = 0

        if rng is not None:
            # numpy interval is different to ``randint``
            return rng.integers(low, high + 1, dtype=np.int32)

        # use standard random for now as it performs better when generating values one at a time
        return random.randint(low, high)

    def _prepareTemplateStrings(self, genTemplate, escapeSpecialMeaning=False):
        """ Prepare list of random numbers needed to generate template in vectorized form

        :param genTemplate: template string to control text generation
        :param escapeSpecialMeaning: if True, requires escape on special meaning chars.
        :returns: tuple containing number of placeholders and vector of random values upper bounds

        The first element of the tuple is the number of placeholders needed to populate the template

        The second elememt is a vector of integer values which determine bounds for random number vector for
        template generation

        Each element of the vector will be used to generate a random number between 0 and the element inclusive,
        which is then used to select words from wordlists etc for template expansion

        `_escapeSpecialMeaning` parameter allows for backwards compatibility with old style syntax while allowing
        for preferred new style template syntax. Specify as True to force escapes for special meanings,.

        """
        retval = []

        escape = False
        use_value = False
        template_len = len(genTemplate)
        num_placeholders = 0

        # in the following code, the construct `(not escape) ^ self._escapeSpecialMeaning` means apply
        # special meaning if either escape is not true or the option `self._escapeSpecialMeaning` is true.
        # This corresponds to the logical xor operation
        for i in range(0, template_len):
            char = genTemplate[i]
            following_char = genTemplate[i + 1] if i + 1 < template_len else None

            if char == '\\':
                escape = True
            elif use_value and ('0' <= char <= '9'):
                # val_index = int(char)
                # retval.append(str(baseValue[val_index]))
                num_placeholders += 1
                use_value = False
            elif (char in self._templateMappings) and (not escape) ^ escapeSpecialMeaning:
                # handle case for ['a','A','k', 'K', 'x', 'X']
                bound, mappingArr = self._templateMappings[char]
                retval.append(bound)
                num_placeholders += 1
                escape = False
            elif (char in self._templateEscapedMappings) and escape:
                # handle case for ['n', 'N', 'w', 'W']
                bound, mappingArr = self._templateEscapedMappings[char]
                retval.append(bound)
                num_placeholders += 1
                escape = False
            elif char == 'v' and escape:
                escape = False
                if following_char is not None and ('0' <= following_char <= '9'):
                    use_value = True
                else:
                    num_placeholders += 1
                    # retval.append(str(baseValue))
            elif char == 'V' and escape:
                # retval.append(str(baseValue))
                num_placeholders += 1
                escape = False
            else:
                # retval.append(char)
                num_placeholders += 1
                escape = False

        if use_value:
            # retval.append(str(baseValue))
            num_placeholders += 1

        return num_placeholders, retval

    def _applyTemplateStringsForTemplate(self, baseValue, genTemplate, placeholders, rnds, escapeSpecialMeaning=False):
        """ Vectorized implementation of template driven text substitution

         Apply substitutions to placeholders using random numbers

        :param baseValue: Pandas series or data frame of base value for applying template
        :param genTemplate: template string to control text generation
        :param placeholders: masked nparray of type np.object_ pre-allocated to hold strings emitted
        :param rnds: masked numpy 2d array of random numbers needed for vectorized generation
        :param escapeSpecialMeaning: if True, requires escape on special meaning chars.
        :returns: placeholders

        The vectorized implementation populates the placeholder Numpy array with the substituted values.

        `_escapeSpecialMeaning` parameter allows for backwards compatibility with old style syntax while allowing
        for preferred new style template syntax. Specify as True to force escapes for special meanings,.

        .. note::
                Both `placeholders` and `rnds` are numpy masked arrays. If there are multiple templates in the template
                generation source template, then this method will be called multiple times with each of
                the distinct templates passed and the `placeholders` and `rnds` arrays masked so that the each call
                will apply the template to rows to which that template applies.

                The template may be the empty string.

        """
        assert baseValue.shape[0] == placeholders.shape[0]
        assert baseValue.shape[0] == rnds.shape[0]

        _cached_values = {}

        regularKeys = self._templateMappings.keys()
        escapedKeys = self._templateEscapedMappings.keys()

        def _get_values_as_np_array():
            """Get baseValue which is pd.Series or Dataframe as a numpy array and cache it"""
            if "np_values" not in _cached_values:
                _cached_values["np_values"] = baseValue.to_numpy()

            return _cached_values["np_values"]

        def _get_values_subelement(elem):
            """Get element from base values as np array and cache it"""
            cache_key = f"v_{elem}"
            if cache_key not in _cached_values:
                np_values = _get_values_as_np_array()
                # element_values = []
                element_values = np.ndarray(np_values.shape[0], dtype=np_values.dtype)

                for x in range(baseValue.shape[0]):
                    # element_values.append(baseValue[x][elem])
                    element_values[x] = baseValue[x][elem]
                _cached_values[cache_key] = element_values

            return _cached_values[cache_key]

        escape = False
        use_value = False
        template_len = len(genTemplate)
        num_placeholders = 0
        rnd_offset = 0

        unmasked_rows = None  # unmasked_rows is None, indicates that all rows are unmasked

        assert isinstance(placeholders, np.ma.MaskedArray), "expecting MaskArray"

        # if template is empty, then nothing needs to be done
        if template_len > 0 and isinstance(placeholders, np.ma.MaskedArray):
            active_rows = ~placeholders.mask
            unmasked_rows = active_rows[:, 0]

            if np.all(active_rows):
                unmasked_rows = None

        # in the following code, the construct `(not escape) ^ self._escapeSpecialMeaning` means apply
        # special meaning if either escape is not true or the option `self._escapeSpecialMeaning` is true.
        # This corresponds to the logical xor operation
        for i in range(0, template_len):
            char = genTemplate[i]
            following_char = genTemplate[i + 1] if i + 1 < template_len else None

            if char == '\\':
                escape = True
            elif use_value and ('0' <= char <= '9'):
                val_index = int(char)
                placeholders[:, num_placeholders] = _get_values_subelement(val_index)
                # placeholders[:, num_placeholders] = pd_base_values.apply(lambda x: str(x[val_index]))
                num_placeholders += 1
                use_value = False
            elif char in regularKeys and (not escape) ^ escapeSpecialMeaning:
                # note vectorized lookup - `rnds[:, rnd_offset]` will get vertical column of
                # random numbers from `rnds` 2d array
                bound, valueMappings = self._templateMappings[char]

                if unmasked_rows is not None:
                    placeholders[unmasked_rows, num_placeholders] = valueMappings[rnds[unmasked_rows, rnd_offset]]
                else:
                    placeholders[:, num_placeholders] = valueMappings[rnds[:, rnd_offset]]

                num_placeholders += 1
                rnd_offset = rnd_offset + 1
                escape = False
                # used for retval.append(_HEX_LOWER[self._getRandomInt(0, 15, rndGenerator)])
            elif char in escapedKeys and escape:
                bound, valueMappings = self._templateEscapedMappings[char]

                if valueMappings is not None:
                    if unmasked_rows is not None:
                        placeholders[unmasked_rows, num_placeholders] = valueMappings[rnds[unmasked_rows, rnd_offset]]
                    else:
                        placeholders[:, num_placeholders] = valueMappings[rnds[:, rnd_offset]]
                else:
                    if unmasked_rows is not None:
                        placeholders[unmasked_rows, num_placeholders] = rnds[unmasked_rows, rnd_offset]
                    else:
                        placeholders[:, num_placeholders] = rnds[:, rnd_offset]
                num_placeholders += 1
                rnd_offset = rnd_offset + 1
                # retval.append(str(self._getRandomInt(0, 255, rndGenerator)))
                escape = False
            elif char == 'v' and escape:
                escape = False
                if following_char is not None and ('0' <= following_char <= '9'):
                    use_value = True
                else:
                    placeholders[:, num_placeholders] = _get_values_as_np_array()
                    num_placeholders += 1
                    # retval.append(str(baseValue))
            elif char == 'V' and escape:
                placeholders[:, num_placeholders] = _get_values_as_np_array()
                # retval.append(str(baseValue))
                num_placeholders += 1
                escape = False
            else:
                placeholders[:, num_placeholders] = char
                # retval.append(char)
                num_placeholders += 1
                escape = False

        if use_value:
            placeholders[:, num_placeholders] = _get_values_as_np_array()
            # retval.append(str(baseValue))
            num_placeholders += 1

        return placeholders

    def classicGenerateText(self, v):
        """entry point to use for classic udfs"""

        pdValues = pd.Series([v])
        results = self.pandasGenerateText(pdValues)
        return results[0]

    def _prepare_random_bounds(self, v):
        """
        Prepare the random bounds for processing of the template expansion

        For each template, we will have a vector of random numbers to generate for expanding the template

        If we have multiple templates, there will be a separate vector of random numbers for each template


        :param v: Pandas series of values passed as base values
        :return: vector of templates chosen, template random bounds (1 for each substitution) and selected
                 random numbers for each row (as numpy array)
        """
        # choose templates
        num_templates = len(self.templates)
        assert num_templates >= 1, "Expecting at least 1 template"

        rng = self.getNPRandomGenerator()

        if num_templates > 1:
            # choose template at random from 0 .. num_templates
            templates_chosen = rng.integers(np.full(v.size, num_templates))
        else:
            # always use template 0
            templates_chosen = np.full(v.size, 0)

        # populate template random numbers
        template_rnd_bounds = np.full((v.size, self._max_rnds_needed), -1)
        masked_template_bounds = np.ma.MaskedArray(template_rnd_bounds, mask=False)

        for i in range(num_templates):
            # assign the
            len_bounds_i = len(self._template_rnd_bounds[i])
            masked_template_bounds[templates_chosen.T == i, 0:len_bounds_i] = self._template_rnd_bounds[i]

        masked_template_bounds[template_rnd_bounds == -1] = np.ma.masked

        r_mask = masked_template_bounds.mask

        template_rnds = template_rnd_bounds.copy()

        template_rnds[~r_mask] = rng.integers(masked_template_bounds[~r_mask])

        return templates_chosen, template_rnd_bounds, template_rnds

    def pandasGenerateText(self, v):
        """ entry point to use for pandas udfs

        Implementation uses vectorized implementation of process

        :param v: Pandas series of values passed as base values
        :return: Pandas series of expanded templates

        """
        # placeholders is numpy array used to hold results
        placeholders = np.full((v.shape[0], self._max_placeholders), '', dtype=np.object_)

        # prepare template selections, bounds, rnd values to drive application of algorithm
        template_choices, template_rnd_bounds, template_rnds = self._prepare_random_bounds(v)
        template_choices_t = template_choices.T

        # create masked arrays, with all elements initially masked
        # as we substitute template expansion, we'll mask and unmask rows corresponding to each template
        # calling the method to substitute the values on the masked placeholders
        masked_placeholders = np.ma.MaskedArray(placeholders, mask=False)
        masked_rnds = np.ma.MaskedArray(template_rnds, mask=False)
        # masked_base_values = np.ma.MaskedArray(baseValues, mask=False)
        masked_matrices = [masked_placeholders, masked_rnds]

        # test logic for template expansion
        for x in range(len(self._templates)):  # pylint: disable=consider-using-enumerate
            masked_placeholders[template_choices_t != x, :] = np.ma.masked
            masked_rnds[template_choices_t != x, :] = np.ma.masked
            # masked_base_values[template_choices_t != x] = np.ma.masked

            # harden mask, preventing modifications
            for m in masked_matrices:
                np.ma.harden_mask(m)

            # expand values into placeholders without affect masked values
            #self._applyTemplateStringsForTemplate(v.to_numpy(dtype=np.object_), #masked_base_values,
            self._applyTemplateStringsForTemplate(v,
                                                  # masked_base_values,
                                                  self._templates[x],
                                                  masked_placeholders,
                                                  masked_rnds,
                                                  escapeSpecialMeaning=self._escapeSpecialMeaning
                                                  )

            # soften and clear mask, allowing modifications
            for m in masked_matrices:
                np.ma.soften_mask(m)
                m.mask = False

        # join strings in placeholders
        output = pd.Series(list(placeholders))
        results = output.apply(lambda placeholder_items: "".join([str(elem) for elem in placeholder_items]))

        return results


class ILText(TextGenerator):  # lgtm [py/missing-equals]
    """ Class to generate Ipsum Lorem text paragraphs, words and sentences

    :param paragraphs: Number of paragraphs to generate. If tuple will generate random number in range
    :param sentences:  Number of sentences to generate. If tuple will generate random number in tuple range
    :param words:  Number of words per sentence to generate. If tuple, will generate random number in tuple range

    """

    def __init__(self, paragraphs=None, sentences=None, words=None, extendedWordList=None):
        """
        Initialize the ILText with text generation parameters
        """
        assert paragraphs is not None or sentences is not None or words is not None, \
            "At least one of the params `paragraphs`, `sentences` or `words` must be specified"

        super().__init__()

        self.paragraphs = self.getAsTupleOrElse(paragraphs, (1, 1), "paragraphs")
        self.words = self.getAsTupleOrElse(words, (2, 12), "words")
        self.sentences = self.getAsTupleOrElse(sentences, (1, 1), "sentences")
        self.wordList = extendedWordList if extendedWordList is not None else _WORDS_LOWER
        self.shape = [self.paragraphs[1], self.sentences[1], self.words[1]]

        # values needed for the text generation
        # numpy uses fixed sizes for strings , so compute whats needed
        self._npWords = np.array(self.wordList)

        self._processStats()
        self._processWordList()

    def _processStats(self):
        """ Compute the stats needed for the text generation """

        vals = [self.paragraphs, self.sentences, self.words]
        self._textGenerationValues = np.array(vals, dtype=self.compactNumpyTypeForValues(vals))
        self._minValues = self._textGenerationValues[:, 0]
        self._maxValues = self._textGenerationValues[:, 1]

        self._meanValues = np.mean(self._textGenerationValues, axis=1)

        # we want to force wider spread of sentence length, so we're not simply computing the std_deviation
        # - but computing a target std_dev that will spread sentence length
        self._stdVals = self._meanValues / 2
        self._stdVals2 = np.std(self._textGenerationValues, axis=1)

    def _processWordList(self):
        """ Set up the word lists"""
        np_words = np.array(self.wordList, np.dtype(np.str_))
        np_capitalized_words = np.char.capitalize(np_words[:])

        all_words = np_words[:]

        self._wordOffsetSize = all_words.size
        self._sentenceEndOffset = all_words.size
        self._paragraphEnd = self._sentenceEndOffset + 1
        self._wordSpaceOffset = self._paragraphEnd + 1
        self._emptyStringOffset = self._wordSpaceOffset + 1

        punctuation = [". ", "\n\n", " ", ""]
        all_words = np.concatenate((all_words, punctuation))

        self._startOfCapitalsOffset = all_words.size
        all_words = np.concatenate((all_words, np_capitalized_words, punctuation))

        # for efficiency, we'll create list of words preceded by spaces - it will reduce memory consumption during join
        # and array manipulation as we dont have to hold offset for space
        self._startOfSpacedWordsOffset = all_words.size

        np_spaced_words = np.array([" " + x for x in self.wordList], np.dtype(np.str_))
        all_words = np.concatenate((all_words, np_spaced_words, punctuation))

        # set up python list of all words so that we dont have to convert between numpy and python representations
        self._allWordsSize = all_words.size
        self._wordsAsPythonStrings = [str(x) for x in all_words]

        # get smallest type that can represent word offset
        self._wordOffsetType = self.compactNumpyTypeForValues([all_words.size * 2 + 10])

    def __repr__(self):
        paras, sentences, words = self.paragraphs, self.sentences, self.words
        wl = self.wordList.__repr__ if self.wordList is not None else "None"
        return f"ILText(paragraphs={paras}, sentences={sentences}, words={words}, wordList={wl})"

    def generateText(self, baseValues, rowCount=1):
        """
        generate text for seed based on configuration parameters.

        As it uses numpy, repeatability is restricted depending on version of the runtime

        :param baseValues: list or array-like list of baseValues
        :param rowCount: number of rows
        :returns: list or Pandas series of generated strings of same size as input seed
        """
        assert baseValues is not None, "`baseValues` param must be specified"
        rng = self.getNPRandomGenerator(forceNewInstance=True)
        word_offset_type = self._wordOffsetType

        stats_shape = [rowCount, self.paragraphs[1], self.sentences[1], 3]

        # determine counts of paragraphs, sentences and words
        stats_array = np.empty(stats_shape, dtype=self._textGenerationValues.dtype)
        para_stats_raw = np.round(rng.normal(self._meanValues, self._stdVals2, size=stats_shape))
        para_stats = np.clip(para_stats_raw, self._minValues, self._maxValues, out=stats_array)

        # replicate paragraphs and sentences from first row of each paragraph through other elememnts
        # this is used to efficiently create mask for manipulating word offsets
        # after this every row will contain :
        # number of good paragraphs, number of good sentences, number of number of good words
        # for current sentence in current paragraph in outer rows collection
        para_stats[:, :, :, 0] = para_stats[:, :, 0, 0, np.newaxis]
        para_stats[:, :, :, 1] = para_stats[:, :, 0, 1, np.newaxis]

        # set up shape for arrays used to process word offsets
        # this will be masked so that we don't waste resources processing invalid paragraphs, sentences and words
        output_shape = (rowCount, self.paragraphs[1], self.sentences[1], self.words[1])

        # compute the masks for paragraphs, sentences, and words

        # get the set of indices for shape  - r = rows, p = paragraphs, s = sentences, w = words
        # the indices will produce a set of rows of values for each dimension
        # the mask is then produced by iterating comparing index with good value
        # for example - if number of good words is 3 and sentence is index array of [0, 1, 2, 3, 4, 5 ,6]
        # then mask is produced via conditions indices <= good_word
        # note value of True means masked for numpy
        r, p, s, w = np.indices(output_shape)

        good_words = para_stats[:, :, :, 2]
        good_paragraphs = para_stats[:, :, :, 0]
        good_sentences = para_stats[:, :, :, 1]

        # build masks in each dimension and `or` them together
        words_mask = (w.T >= good_words.T).T
        para_mask = (p.T >= good_paragraphs.T).T
        sentences_mask = (s.T >= good_sentences.T).T
        final_mask = words_mask | para_mask | sentences_mask

        word_offsets = np.full(output_shape, dtype=word_offset_type, fill_value=self._emptyStringOffset)
        masked_offsets = np.ma.MaskedArray(word_offsets, mask=final_mask)

        # note numpy random differs from standard random in that it never produces upper bound
        masked_offsets[~masked_offsets.mask] = rng.integers(self._wordOffsetSize,
                                                            size=output_shape,
                                                            dtype=self._wordOffsetType)[~masked_offsets.mask]

        # hardening a mask prevents masked values from being changed
        np.ma.harden_mask(masked_offsets)
        masked_offsets[:, :, :, 0] = masked_offsets[:, :, :, 0] + self._startOfCapitalsOffset
        masked_offsets[:, :, :, 1:] = masked_offsets[:, :, :, 1:] + self._startOfSpacedWordsOffset
        np.ma.soften_mask(masked_offsets)

        # add period to every sentence
        # we'll replicate column 0 in order to preserve the mask and fill unmasked entries
        # with the sentence end offset
        new_word_offsets = masked_offsets[:, :, :, 0][:]
        new_col = new_word_offsets[:, :, :, np.newaxis]
        terminated_word_offsets = np.ma.concatenate((masked_offsets, new_col), axis=3)
        new_column = terminated_word_offsets[:, :, :, -1]
        new_column[~new_column.mask] = self._sentenceEndOffset

        # reshape to paragraphs
        shape = terminated_word_offsets.shape
        paragraph_offsets = terminated_word_offsets.reshape((rowCount, shape[1], shape[2] * shape[3]))

        if self.paragraphs[1] > 1:
            # add paragraph terminator to every paragraph
            # we'll take a copy of the first column so as to preserve masking
            # i.e if first word of first sentence is masked, then end marker position should be masked
            new_word_offsets = paragraph_offsets[:, :, 0][:]
            new_col = new_word_offsets[:, :, np.newaxis]
            terminated_paragraph_offsets = np.ma.concatenate((paragraph_offsets, new_col), axis=2)

            # set the paragraph end marker on all paragraphs except last
            # new_masked_elements = terminated_paragraph_offsets[:,:,-1]
            new_column = terminated_paragraph_offsets[:, :, -1]
            new_column[~new_column.mask] = self._paragraphEnd
        else:
            terminated_paragraph_offsets = paragraph_offsets

        # reshape to rows containing word offset sequences. We'll join using pandas apply to avoid
        # memory issues with fixed strings in numpy
        shape = terminated_paragraph_offsets.shape
        terminated_paragraph_offsets = terminated_paragraph_offsets.reshape((rowCount, shape[1] * shape[2]))

        final_data = terminated_paragraph_offsets.filled(fill_value=self._emptyStringOffset)

        # its faster to manipulate text in data frames as numpy strings are fixed length
        all_python_words = self._wordsAsPythonStrings

        base_results = pd.DataFrame(final_data)

        # build our lambda expression, copying point to word list locally for efficiency
        empty_string_offsets = [self._emptyStringOffset, self._emptyStringOffset + self._startOfSpacedWordsOffset]
        mk_str_fn = lambda x: ("".join([all_python_words[x1] for x1 in x if x1 not in empty_string_offsets])).strip()
        # mk_str_fn = lambda x: ("".join([all_python_words[x1] for x1 in x ]))

        # ... and execute it
        results = base_results.apply(mk_str_fn, axis=1)
        return results

    def classicGenerateText(self, v):
        """
        classic udf entry point for text generation

        :param v: base value to control generation of random numbers
        """
        return self.generateText([v], 1)[0]

    def pandasGenerateText(self, v):
        """
        pandas udf entry point for text generation

        :param v: pandas series of base values for random text generation
        :returns: Pandas series of generated strings
        """
        rows = v.to_numpy()
        results = self.generateText(rows, rows.size)
        return pd.Series(results)



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the Beta statistical distributions related classes

"""

import pyspark.sql.functions as F
from pyspark.sql.types import FloatType

import numpy as np
import pandas as pd

from .data_distribution import DataDistribution


class Beta(DataDistribution):
    """ Specify that random samples should be drawn from the Beta distribution parameterized by alpha and beta

    :param alpha: value for alpha parameter - float, int or other numeric value, greater than 0
    :param beta: value for beta parameter - float, int or other numeric value, greater than 0

    See https://en.wikipedia.org/wiki/Beta_distribution

    By default the Beta distribution produces values between 0 and 1 so no scaling is needed
    """

    def __init__(self, alpha=None, beta=None):
        DataDistribution.__init__(self)

        assert type(alpha) in [float, int, np.float64, np.int32, np.int64], "alpha must be int-like or float-like"
        assert type(beta) in [float, int, np.float64, np.int32, np.int64], "beta must be int-like or float-like"
        self._alpha = alpha
        self._beta = beta

    @property
    def alpha(self):
        """ Return alpha parameter."""
        return self._alpha

    @property
    def beta(self):
        """ Return beta parameter."""
        return self._beta

    def __str__(self):
        """ Return string representation of object"""
        return f"BetaDistribution(alpha={self._alpha}, beta={self._beta}, randomSeed={self.randomSeed})"

    @staticmethod
    def beta_func(alpha_series: pd.Series, beta_series: pd.Series, random_seed: pd.Series) -> pd.Series:
        """ Generate sample of beta distribution using pandas / numpy

        :param alpha_series: value for alpha parameter as Pandas Series
        :param beta_series: value for beta parameter as Pandas Series
        :param random_seed: value for randomSeed parameter as Pandas Series

        :return: random samples from distribution scaled to values between 0 and 1

        """
        alpha = alpha_series.to_numpy()
        beta = beta_series.to_numpy()
        random_seed = random_seed.to_numpy()[0]

        rng = DataDistribution.get_np_random_generator(random_seed)

        results = rng.beta(alpha, beta)
        return pd.Series(results)

    def generateNormalizedDistributionSample(self):
        """ Generate sample of data for distribution

        :return: random samples from distribution scaled to values between 0 and 1
        """
        beta_sample = F.pandas_udf(self.beta_func, returnType=FloatType()).asNondeterministic()

        newDef = beta_sample(F.lit(self._alpha),
                             F.lit(self._beta),
                             F.lit(self.randomSeed) if self.randomSeed is not None else F.lit(-1))
        return newDef



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the Exponential statistical distributions related classes

"""

import numpy as np
import pandas as pd

import pyspark.sql.functions as F
from pyspark.sql.types import FloatType

from .data_distribution import DataDistribution


class Exponential(DataDistribution):
    """ Specify that random samples should be drawn from the exponential distribution parameterized by rate

    :param rate: value for rate parameter - float, int or other numeric value, greater than 0

    See https://en.wikipedia.org/wiki/Exponential_distribution

    Scaling is performed to normalize values between 0 and 1
    """

    def __init__(self, rate=None):
        DataDistribution.__init__(self)
        self._rate = rate

    def __str__(self):
        """ Return string representation"""
        return f"ExponentialDistribution(rate={self.rate}, randomSeed={self.randomSeed})"

    @staticmethod
    def exponential_func(scale_series: pd.Series, random_seed: pd.Series) -> pd.Series:
        """ Generate sample of exponential distribution using pandas / numpy

        :param scale_series: value for scale parameter as Pandas Series
        :param random_seed: value for randomSeed parameter as Pandas Series
        :return: random samples from distribution scaled to values between 0 and 1

        """
        scale_param = scale_series.to_numpy()
        random_seed = random_seed.to_numpy()[0]

        rng = DataDistribution.get_np_random_generator(random_seed)

        results = rng.exponential(scale_param)

        # scale the results
        amin = np.amin(results) * 1.0
        amax = np.amax(results) * 1.0

        adjusted_results = results - amin

        scaling_factor = amax - amin

        results2 = adjusted_results / scaling_factor

        return pd.Series(results2)

    @property
    def rate(self):
        """ Return rate parameter"""
        return self._rate

    @property
    def scale(self):
        """ Return scale implicit parameter. Scale is 1/rate"""
        return 1.0 / self._rate

    def generateNormalizedDistributionSample(self):
        """ Generate sample of data for distribution

        :return: random samples from distribution scaled to values between 0 and 1
        """
        exponential_sample = F.pandas_udf(self.exponential_func, returnType=FloatType()).asNondeterministic()

        # scala formulation uses scale = 1/rate
        newDef = exponential_sample(F.lit(1.0 / self._rate),
                             F.lit(self.randomSeed) if self.randomSeed is not None else F.lit(-1.0))
        return newDef



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This module defines the ``DataAnalyzer`` class.

This code is experimental and both APIs and code generated is liable to change in future versions.
"""
from pyspark.sql.types import LongType, FloatType, IntegerType, StringType, DoubleType, BooleanType, ShortType, \
    TimestampType, DateType, DecimalType, ByteType, BinaryType, StructType, ArrayType, DataType

import pyspark.sql as ssql
import pyspark.sql.functions as F

from .utils import strip_margins
from .spark_singleton import SparkSingleton


class DataAnalyzer:
    """This class is used to analyze an existing data set to assist in generating a test data set with similar
    characteristics, and to generate code from existing schemas and data

    :param df: Spark dataframe to analyze
    :param sparkSession: Spark session instance to use when performing spark operations

    .. warning::
       Experimental

    """
    _DEFAULT_GENERATED_NAME = "synthetic_data"

    _GENERATED_COMMENT = strip_margins("""
                        |# Code snippet generated with Databricks Labs Data Generator (`dbldatagen`) DataAnalyzer class
                        |# Install with `pip install dbldatagen` or in notebook with `%pip install dbldatagen`
                        |# See the following resources for more details:
                        |#
                        |#   Getting Started - [https://databrickslabs.github.io/dbldatagen/public_docs/APIDOCS.html]
                        |#   Github project - [https://github.com/databrickslabs/dbldatagen]
                        |#""", '|')

    _GENERATED_FROM_SCHEMA_COMMENT = strip_margins("""
                        |# Column definitions are stubs only - modify to generate correct data  
                        |#""", '|')

    def __init__(self, df=None, sparkSession=None):
        """ Constructor:
        :param df: Dataframe to analyze
        :param sparkSession: Spark session to use
        """
        assert df is not None, "dataframe must be supplied"

        self._df = df

        if sparkSession is None:
            sparkSession = SparkSingleton.getLocalInstance()

        self._sparkSession = sparkSession
        self._dataSummary = None

    def _displayRow(self, row):
        """Display details for row"""
        results = []
        row_key_pairs = row.asDict()
        for x in row_key_pairs:
            results.append(f"{x}: {row[x]}")

        return ", ".join(results)

    def _addMeasureToSummary(self, measureName, summaryExpr="''", fieldExprs=None, dfData=None, rowLimit=1,
                             dfSummary=None):
        """ Add a measure to the summary dataframe

        :param measureName: Name of measure
        :param summaryExpr: Summary expression
        :param fieldExprs: list of field expressions (or generator)
        :param dfData: Source data df - data being summarized
        :param rowLimit: Number of rows to get for measure
        :param dfSummary: Summary df
        :return: dfSummary with new measure added
        """
        assert dfData is not None, "source data dataframe must be supplied"
        assert measureName is not None and len(measureName) > 0, "invalid measure name"

        # add measure name and measure summary
        exprs = [f"'{measureName}' as measure_", f"string({summaryExpr}) as summary_"]

        # add measures for fields
        exprs.extend(fieldExprs)

        if dfSummary is not None:
            dfResult = dfSummary.union(dfData.selectExpr(*exprs).limit(rowLimit))
        else:
            dfResult = dfData.selectExpr(*exprs).limit(rowLimit)

        return dfResult

    def summarizeToDF(self):
        """ Generate summary analysis of data set as dataframe

        :return: Summary results as dataframe

        The resulting dataframe can be displayed with the ``display`` function in a notebook environment
        or with the ``show`` method.

        The output is also used in code generation  to generate more accurate code.
        """
        self._df.cache().createOrReplaceTempView("data_analysis_summary")

        total_count = self._df.count() * 1.0

        dtypes = self._df.dtypes

        # schema information
        dfDataSummary = self._addMeasureToSummary(
            'schema',
            summaryExpr=f"""to_json(named_struct('column_count', {len(dtypes)}))""",
            fieldExprs=[f"'{dtype[1]}' as {dtype[0]}" for dtype in dtypes],
            dfData=self._df)

        # count
        dfDataSummary = self._addMeasureToSummary(
            'count',
            summaryExpr=f"{total_count}",
            fieldExprs=[f"string(count({dtype[0]})) as {dtype[0]}" for dtype in dtypes],
            dfData=self._df,
            dfSummary=dfDataSummary)

        dfDataSummary = self._addMeasureToSummary(
            'null_probability',
            fieldExprs=[f"""string( round( ({total_count} - count({dtype[0]})) /{total_count}, 2)) as {dtype[0]}"""
                        for dtype in dtypes],
            dfData=self._df,
            dfSummary=dfDataSummary)

        # distinct count
        dfDataSummary = self._addMeasureToSummary(
            'distinct_count',
            summaryExpr="count(distinct *)",
            fieldExprs=[f"string(count(distinct {dtype[0]})) as {dtype[0]}" for dtype in dtypes],
            dfData=self._df,
            dfSummary=dfDataSummary)

        # min
        dfDataSummary = self._addMeasureToSummary(
            'min',
            fieldExprs=[f"string(min({dtype[0]})) as {dtype[0]}" for dtype in dtypes],
            dfData=self._df,
            dfSummary=dfDataSummary)

        dfDataSummary = self._addMeasureToSummary(
            'max',
            fieldExprs=[f"string(max({dtype[0]})) as {dtype[0]}" for dtype in dtypes],
            dfData=self._df,
            dfSummary=dfDataSummary)

        descriptionDf = self._df.describe().where("summary in ('mean', 'stddev')")
        describeData = descriptionDf.collect()

        for row in describeData:
            measure = row['summary']

            values = {k[0]: '' for k in dtypes}

            row_key_pairs = row.asDict()
            for k1 in row_key_pairs:
                values[k1] = str(row[k1])

            dfDataSummary = self._addMeasureToSummary(
                measure,
                fieldExprs=[f"'{values[dtype[0]]}'" for dtype in dtypes],
                dfData=self._df,
                dfSummary=dfDataSummary)

        # string characteristics for strings and string representation of other values
        dfDataSummary = self._addMeasureToSummary(
            'print_len_min',
            fieldExprs=[f"min(length(string({dtype[0]}))) as {dtype[0]}" for dtype in dtypes],
            dfData=self._df,
            dfSummary=dfDataSummary)

        dfDataSummary = self._addMeasureToSummary(
            'print_len_max',
            fieldExprs=[f"max(length(string({dtype[0]}))) as {dtype[0]}" for dtype in dtypes],
            dfData=self._df,
            dfSummary=dfDataSummary)

        return dfDataSummary

    def summarize(self, suppressOutput=False):
        """ Generate summary analysis of data set and return / print summary results

        :param suppressOutput:  If False, prints results to console also
        :return: Summary results as string
        """
        dfSummary = self.summarizeToDF()

        results = [
            "Data set summary",
            "================"
        ]

        for r in dfSummary.collect():
            results.append(self._displayRow(r))

        summary = "\n".join([str(x) for x in results])

        if not suppressOutput:
            print(summary)

        return summary

    @classmethod
    def _valueFromSummary(cls, dataSummary, colName, measure, defaultValue):
        """ Get value from data summary

        :param dataSummary: Data summary to search, optional
        :param colName: Column name of column to get value for
        :param measure: Measure name of measure to get value for
        :param defaultValue: Default value if any other argument is not specified or value could not be found in
                             data summary
        :return: Value from lookup or `defaultValue` if not found
        """
        if dataSummary is not None and colName is not None and measure is not None:
            if measure in dataSummary:
                measureValues = dataSummary[measure]

                if colName in measureValues:
                    return measureValues[colName]

        # return default value if value could not be looked up or found
        return defaultValue

    @classmethod
    def _generatorDefaultAttributesFromType(cls, sqlType, colName=None, dataSummary=None, sourceDf=None):
        """ Generate default set of attributes for each data type

        :param sqlType: Instance of `pyspark.sql.types.DataType`
        :param colName: Name of column being generated
        :param dataSummary: Map of maps of attributes from data summary, optional
        :param sourceDf: Source dataframe to retrieve attributes of real data, optional
        :return: Attribute string for supplied sqlType

        When generating code from a schema, we have no data heuristics to determine how data should be generated,
        so goal is to just generate code that produces some data.

        Users are expected to modify the generated code to their needs.
        """
        assert isinstance(sqlType, DataType)

        if sqlType == StringType():
            result = """template=r'\\\\w'"""
        elif sqlType in [IntegerType(), LongType()]:
            minValue = cls._valueFromSummary(dataSummary, colName, "min", defaultValue=0)
            maxValue = cls._valueFromSummary(dataSummary, colName, "max", defaultValue=1000000)
            result = f"""minValue={minValue}, maxValue={maxValue}"""
        elif sqlType == ByteType():
            minValue = cls._valueFromSummary(dataSummary, colName, "min", defaultValue=0)
            maxValue = cls._valueFromSummary(dataSummary, colName, "max", defaultValue=127)
            result = f"""minValue={minValue}, maxValue={maxValue}"""
        elif sqlType == ShortType():
            minValue = cls._valueFromSummary(dataSummary, colName, "min", defaultValue=0)
            maxValue = cls._valueFromSummary(dataSummary, colName, "max", defaultValue=32767)
            result = f"""minValue={minValue}, maxValue={maxValue}"""
        elif sqlType == BooleanType():
            result = """expr='id % 2 = 1'"""
        elif sqlType == DateType():
            result = """expr='current_date()'"""
        elif isinstance(sqlType, DecimalType):
            minValue = cls._valueFromSummary(dataSummary, colName, "min", defaultValue=0)
            maxValue = cls._valueFromSummary(dataSummary, colName, "max", defaultValue=1000)
            result = f"""minValue={minValue}, maxValue={maxValue}"""
        elif sqlType in [FloatType(), DoubleType()]:
            minValue = cls._valueFromSummary(dataSummary, colName, "min", defaultValue=0.0)
            maxValue = cls._valueFromSummary(dataSummary, colName, "max", defaultValue=1000000.0)
            result = f"""minValue={minValue}, maxValue={maxValue}, step=0.1"""
        elif sqlType == TimestampType():
            result = """begin="2020-01-01 01:00:00", end="2020-12-31 23:59:00", interval="1 minute" """
        elif sqlType == BinaryType():
            result = """expr="cast('dbldatagen generated synthetic data' as binary)" """
        else:
            result = """expr='null'"""

        percentNullsValue = float(cls._valueFromSummary(dataSummary, colName, "null_probability", defaultValue=0.0))

        if percentNullsValue > 0.0:
            result = result + f", percentNulls={percentNullsValue}"

        return result

    @classmethod
    def _scriptDataGeneratorCode(cls, schema, dataSummary=None, sourceDf=None, suppressOutput=False, name=None):
        """
        Generate outline data generator code from an existing dataframe

        This will generate a data generator spec from an existing dataframe. The resulting code
        can be used to generate a data generation specification.

        Note at this point in time, the code generated is stub code only.
        For most uses, it will require further modification - however it provides a starting point
        for generation of the specification for a given data set.

        The dataframe to be analyzed is the dataframe passed to the constructor of the DataAnalyzer object.

        :param schema: Pyspark schema - i.e manually constructed StructType or return value from `dataframe.schema`
        :param dataSummary: Map of maps of attributes from data summary, optional
        :param sourceDf: Source dataframe to retrieve attributes of real data, optional
        :param suppressOutput: Suppress printing of generated code if True
        :param name: Optional name for data generator
        :return: String containing skeleton code

        """
        assert isinstance(schema, StructType), "expecting valid Pyspark Schema"

        stmts = []

        if name is None:
            name = cls._DEFAULT_GENERATED_NAME

        stmts.append(cls._GENERATED_COMMENT)

        stmts.append("import dbldatagen as dg")
        stmts.append("import pyspark.sql.types")

        stmts.append(cls._GENERATED_FROM_SCHEMA_COMMENT)

        stmts.append(strip_margins(
            f"""generation_spec = (
                                    |    dg.DataGenerator(sparkSession=spark, 
                                    |                     name='{name}', 
                                    |                     rows=100000,
                                    |                     random=True,
                                    |                     )""",
            '|'))

        indent = "    "
        for fld in schema.fields:
            col_name = fld.name
            col_type = fld.dataType.simpleString()

            if isinstance(fld.dataType, ArrayType):
                col_type = fld.dataType.elementType.simpleString()
                field_attributes = cls._generatorDefaultAttributesFromType(fld.dataType.elementType)  # no data look up
                array_attributes = """structType='array', numFeatures=(2,6)"""
                name_and_type = f"""'{col_name}', '{col_type}'"""
                stmts.append(indent + f""".withColumn({name_and_type}, {field_attributes}, {array_attributes})""")
            else:
                field_attributes = cls._generatorDefaultAttributesFromType(fld.dataType,
                                                                           colName=col_name,
                                                                           dataSummary=dataSummary,
                                                                           sourceDf=sourceDf)
                stmts.append(indent + f""".withColumn('{col_name}', '{col_type}', {field_attributes})""")
        stmts.append(indent + ")")

        if not suppressOutput:
            for line in stmts:
                print(line)

        return "\n".join(stmts)

    @classmethod
    def scriptDataGeneratorFromSchema(cls, schema, suppressOutput=False, name=None):
        """
        Generate outline data generator code from an existing dataframe

        This will generate a data generator spec from an existing dataframe. The resulting code
        can be used to generate a data generation specification.

        Note at this point in time, the code generated is stub code only.
        For most uses, it will require further modification - however it provides a starting point
        for generation of the specification for a given data set.

        The dataframe to be analyzed is the dataframe passed to the constructor of the DataAnalyzer object.

        :param schema: Pyspark schema - i.e manually constructed StructType or return value from `dataframe.schema`
        :param suppressOutput: Suppress printing of generated code if True
        :param name: Optional name for data generator
        :return: String containing skeleton code

        """
        return cls._scriptDataGeneratorCode(schema,
                                            suppressOutput=suppressOutput,
                                            name=name)

    def scriptDataGeneratorFromData(self, suppressOutput=False, name=None):
        """
        Generate outline data generator code from an existing dataframe

        This will generate a data generator spec from an existing dataframe. The resulting code
        can be used to generate a data generation specification.

        Note at this point in time, the code generated is stub code only.
        For most uses, it will require further modification - however it provides a starting point
        for generation of the specification for a given data set

        The dataframe to be analyzed is the Spark dataframe passed to the constructor of the DataAnalyzer object

        :param suppressOutput: Suppress printing of generated code if True
        :param name: Optional name for data generator
        :return: String containing skeleton code

        """
        assert self._df is not None
        assert type(self._df) is ssql.DataFrame, "sourceDf must be a valid Pyspark dataframe"

        if self._dataSummary is None:
            df_summary = self.summarizeToDF()

            self._dataSummary = {}
            for row in df_summary.collect():
                row_key_pairs = row.asDict()
                self._dataSummary[row['measure_']] = row_key_pairs

        return self._scriptDataGeneratorCode(self._df.schema,
                                             suppressOutput=suppressOutput,
                                             name=name,
                                             dataSummary=self._dataSummary,
                                             sourceDf=self._df)



# See the License for the specific language governing permissions and
# limitations under the License.
#
import itertools

from pyspark.sql.types import StringType, DateType, TimestampType


class ColumnGeneratorBuilder:
    """ Helper class to build functional column generators of specific forms"""

    @classmethod
    def _mkList(cls, x):
        """
        Makes a list of the supplied object instance if it is not already a list
        :param x: object to process
        :returns: Returns list of supplied object if it is not already a list, otherwise simply returns the object"""
        return [x] if type(x) is not list else x

    @classmethod
    def _lastElement(cls, x):
        """ Gets the last element, if the object is a list otherwise returns the object itself"""
        return x[-1] if type(x) is list else x

    @classmethod
    def _mkCdfProbabilities(cls, weights):
        """ make cumulative distribution function probabilities for each value in values list

        a cumulative distribution function for discrete values can uses
        a  table of cumulative probabilities to evaluate different expressions
        based on the cumulative  probability of each value occurring

        The probabilities can be computed from the weights using the code::

           weight_interval = 1.0 / total_weights
           probs = [x * weight_interval for x in weights]
           return reduce((lambda x, y: cls._mk_list(x) + [cls._last_element(x) + y]), probs)

        but Python provides built-in methods (itertools.accumulate) to compute the accumulated weights
        and dividing each resulting accumulated weight by the sum of the total weights will give the cumulative
        probabilities.

        For large data sets (relative to number of values), tests verify that the resulting distribution is
        within 10% of the expected distribution when using Spark SQL Rand() as a uniform random number generator.
        In testing, test data sets of size 3000 * number_of_values rows produce a distribution within 10% of expected ,
        while datasets of size 10,000 x `number of values` gives a repeated
        distribution within 5% of expected distribution.

        Example code to be generated (pseudo code)::

           # given values value1 .. valueN, prob1 to probN
           # cdf_probabilities = [ prob1, prob1+prob2, ... prob1+prob2+prob3 .. +probN]
           # this will then be used as follows

           prob_occurrence = uniform_rand(0.0, 1.0)
           if prob_occurrence <= cdf_prob_value1 then value1
           elif prob_occurrence <= cdf_prob_value2 then value2
           ...
           prob_occurrence <= cdf_prob_valueN then valueN
           else valueN

        """
        total_weights = sum(weights)
        return list(map(lambda x: x / total_weights, itertools.accumulate(weights)))

    @classmethod
    def mkExprChoicesFn(cls, values, weights, seed_column, datatype):
        """ Create SQL expression to compute the weighted values expression

        build an expression of the form::

           case
              when rnd_column <= weight1 then value1
              when rnd_column <= weight2 then value2
              ...
              when rnd_column <= weightN then  valueN
              else valueN
           end

        based on computed probability distribution for values.

        In Python 3.6 onwards, we could use the choices function but this python version is not
        guaranteed on all Databricks distributions

        :param values: list of values
        :param weights: list of weights
        :param seed_column: base column for expression
        :param datatype: data type of function return value

        """
        cdf_probs = cls._mkCdfProbabilities(weights)

        output = [" CASE "]

        conditions = zip(values, cdf_probs)

        for v, cdf in conditions:
            # TODO(alex): single quotes needs to be escaped
            if isinstance(datatype, (StringType, DateType, TimestampType)):
                output.append(f" when {seed_column} <= {cdf} then '{v}' ")
            else:
                output.append(f" when {seed_column} <= {cdf} then {v} ")

        # TODO(alex): single quotes needs to be escaped
        output.append("else '{values[-1]}'")
        output.append("end")

        retval = "\n".join(output)

        return retval



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the Normal / Gaussian statistical distributions related classes

"""

import numpy as np
import pandas as pd

import pyspark.sql.functions as F
from pyspark.sql.types import FloatType

from .data_distribution import DataDistribution


class Normal(DataDistribution):
    def __init__(self, mean, stddev):
        ''' Specify that data should follow normal distribution

        :param mean: mean of distribution
        :param stddev: standard deviation of distribution
        '''
        DataDistribution.__init__(self)
        self.mean = mean if mean is not None else 0.0
        self.stddev = stddev if stddev is not None else 1.0

    @staticmethod
    def normal_func(mean_series: pd.Series, std_dev_series: pd.Series, random_seed: pd.Series) -> pd.Series:
        """ Pandas / Numpy based function to generate normal / gaussian samples

        :param mean_series: pandas series of mean values
        :param std_dev_series: pandas series of standard deviation values
        :param random_seed:  pandas series of random seed values

        :return: Samples scaled from 0 .. 1
        """

        mean = mean_series.to_numpy()
        std_dev = std_dev_series.to_numpy()
        random_seed = random_seed.to_numpy()[0]

        rng = DataDistribution.get_np_random_generator(random_seed)

        results = rng.normal(mean, std_dev)

        # scale the results
        amin = np.amin(results) * 1.0
        amax = np.amax(results) * 1.0

        adjusted_results = results - amin

        scaling_factor = amax - amin

        results2 = adjusted_results / scaling_factor
        return pd.Series(results2)

    def generateNormalizedDistributionSample(self):
        """ Generate sample of data for distribution

        :return: random samples from distribution scaled to values between 0 and 1
        """
        normal_sample = F.pandas_udf(self.normal_func, returnType=FloatType()).asNondeterministic()

        # scala formulation uses scale = 1/rate
        newDef = normal_sample(F.lit(self.mean),
                               F.lit(self.stddev),
                             F.lit(self.randomSeed) if self.randomSeed is not None else F.lit(-1.0))
        return newDef

    def __str__(self):
        """ Return string representation of object """
        return f"NormalDistribution( mean={self.mean}, stddev={self.stddev}, randomSeed={self.randomSeed})"

    @classmethod
    def standardNormal(cls):
        """ return instance of standard normal distribution """
        return Normal(mean=0.0, stddev=1.0)



# See the License for the specific language governing permissions and
# limitations under the License.
#
import math
from datetime import datetime, timedelta

from .datarange import DataRange
from .utils import parse_time_interval


class DateRange(DataRange):
    """Class to represent Date range

    The date range will represented internally using `datetime` for `start` and `end`, and `timedelta` for `interval`

    When computing ranges for purposes of the sequences, the maximum value will be adjusted to the
    nearest whole multiple of the interval that is before the `end` value.

    When converting from a string, datetime is assumed to use local timezone unless specified as part of the format
    in keeping with the python `datetime` handling of datetime instances that do not specify a timezone

    :param begin: start of date range as python `datetime` object. If specified as string, converted to datetime
    :param end: end of date range as python `datetime` object. If specified as string, converted to datetime
    :param interval: interval of date range as python `timedelta` object.
                     Note parsing format for interval uses standard timedelta parsing not
                     the `datetime_format` string
    :param datetime_format: format for conversion of strings to datetime objects

    """

    DEFAULT_UTC_TS_FORMAT = "%Y-%m-%d %H:%M:%S"
    DEFAULT_DATE_FORMAT = "%Y-%m-%d"

    DEFAULT_START_TIMESTAMP = datetime(year=datetime.now().year - 1, month=1, day=1)
    DEFAULT_END_TIMESTAMP = datetime(year=datetime.now().year - 1, month=12, day=31, hour=23, minute=59, second=59)
    DEFAULT_START_DATE = DEFAULT_START_TIMESTAMP.date()
    DEFAULT_END_DATE = DEFAULT_END_TIMESTAMP.date()
    DEFAULT_START_DATE_TIMESTAMP = DEFAULT_START_TIMESTAMP.replace(hour=0, minute=0, second=0)
    DEFAULT_END_DATE_TIMESTAMP = DEFAULT_END_TIMESTAMP.replace(hour=0, minute=0, second=0)

    # todo: deduce format from begin and end params

    def __init__(self, begin, end, interval=None, datetime_format=DEFAULT_UTC_TS_FORMAT):
        assert begin is not None, "`begin` must be specified"
        assert end is not None, "`end` must be specified"

        self.begin = begin if not isinstance(begin, str) else self._datetime_from_string(begin, datetime_format)
        self.end = end if not isinstance(end, str) else self._datetime_from_string(end, datetime_format)
        self.interval = interval if not isinstance(interval, str) else self._timedelta_from_string(interval)

        self.minValue = self.begin.timestamp()

        self.maxValue = (self.minValue + self.interval.total_seconds()
                         * self.computeTimestampIntervals(self.begin, self.end, self.interval))
        self.step = self.interval.total_seconds()

    @classmethod
    def _datetime_from_string(cls, date_str, date_format):
        """convert string to Python DateTime object using format"""
        result = datetime.strptime(date_str, date_format)
        return result

    @classmethod
    def _timedelta_from_string(cls, interval):
        return cls.parseInterval(interval)

    @classmethod
    def parseInterval(cls, interval_str):
        """Parse interval from string"""
        assert interval_str is not None, "`interval_str` must be specified"
        return parse_time_interval(interval_str)

    @classmethod
    def _getDateTime(cls, dt, datetime_format, default_value):
        if isinstance(dt, str):
            effective_dt = cls._datetime_from_string(dt, datetime_format)
        elif dt is None:
            effective_dt = default_value
        else:
            effective_dt = dt
        return effective_dt

    @classmethod
    def _getInterval(cls, interval, default_value):
        if isinstance(interval, str):
            effective_interval = parse_time_interval(interval)
        elif interval is None:
            effective_interval = default_value
        else:
            effective_interval = interval
        return effective_interval

    @classmethod
    def computeDateRange(cls, begin, end, interval, unique_values):
        effective_interval = cls._getInterval(interval, timedelta(days=1))
        effective_end = cls._getDateTime(end, DateRange.DEFAULT_DATE_FORMAT,
                                         cls.DEFAULT_END_DATE_TIMESTAMP)
        effective_begin = cls._getDateTime(begin, DateRange.DEFAULT_DATE_FORMAT, cls.DEFAULT_START_DATE_TIMESTAMP)

        if unique_values is not None:
            assert type(unique_values) is int, "unique_values must be integer"
            assert unique_values >= 1, "unique_values must be positive integer"

            effective_begin = effective_end - effective_interval * (unique_values - 1)

        result = DateRange(effective_begin, effective_end, effective_interval)
        return result

    @classmethod
    def computeTimestampRange(cls, begin, end, interval, unique_values):

        effective_interval = cls._getInterval(interval, timedelta(days=1))
        effective_end = cls._getDateTime(end, DateRange.DEFAULT_UTC_TS_FORMAT, cls.DEFAULT_END_TIMESTAMP)
        effective_begin = cls._getDateTime(begin, DateRange.DEFAULT_UTC_TS_FORMAT, cls.DEFAULT_START_TIMESTAMP)

        if unique_values is not None:
            assert type(unique_values) is int, "unique_values must be integer"
            assert unique_values >= 1, "unique_values must be positive integer"
            effective_begin = effective_end - effective_interval * (unique_values - 1)
        result = DateRange(effective_begin, effective_end, effective_interval)
        return result

    def __str__(self):
        """ create string representation of date range"""
        return f"DateRange({self.begin},{self.end},{self.interval} == {self.minValue}, {self.maxValue}, {self.step})"

    def computeTimestampIntervals(self, start, end, interval):
        """ Compute number of intervals between start and end date """
        assert type(start) is datetime, "Expecting start as type datetime.datetime"
        assert type(end) is datetime, "Expecting end as type datetime.datetime"
        assert type(interval) is timedelta, "Expecting interval as type datetime.timedelta"
        i1 = end - start
        ni1 = i1 / interval
        return math.floor(ni1)

    def isFullyPopulated(self):
        """Check if minValue, maxValue and step are specified """
        return self.minValue is not None and self.maxValue is not None and self.step is not None

    def adjustForColumnDatatype(self, ctype):
        """ adjust the range for the column output type

        :param ctype: Spark SQL data type for column

        """
        pass

    def getDiscreteRange(self):
        """ Divide continuous range into discrete intervals

            Note does not modify range object.

            :returns: range from minValue to maxValue

        """
        return (self.maxValue - self.minValue) * float(1.0 / self.step)

    def isEmpty(self):
        """Check if object is empty (i.e all instance vars of note are `None`)"""
        return self.begin is None and self.end is None and self.interval is None

    def getContinuousRange(self):
        """Convert range to continuous range"""
        return (self.maxValue - self.minValue) * float(1.0)

    def getScale(self):
        """Get scale of range"""
        return 0



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This module defines the version information for the test data generator library

Management of version identifiers for releases uses the `bumpversion` python package
which supports automatic modification of files when the version labels need to be modified

See: https://pypi.org/project/bumpversion/

Note the use of `get_version` for method name to conform with bumpversion conventions
"""

from collections import namedtuple
import re
import logging

VersionInfo = namedtuple('VersionInfo', ['major', 'minor', 'patch', 'release', 'build'])


def get_version(version):
    """ Get version info object for library.

    :param version: version string to parse for version information

    Layout of version string must be compatible with `bump` package"""
    r = re.compile(r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+){0,1}(?P<release>\D*)(?P<build>\d*)')
    major, minor, patch, release, build = r.match(version).groups()
    version_info = VersionInfo(major, minor, patch, release, build)
    logger = logging.getLogger(__name__)
    logger.info("Version : %s", version_info)
    return version_info


__version__ = "0.3.6"  # DO NOT EDIT THIS DIRECTLY!  It is managed by bumpversion
__version_info__ = get_version(__version__)


def _get_spark_version(sparkVersion):
    try:
        r = re.compile(r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?P<release>.*)')
        major, minor, patch, release = r.match(sparkVersion).groups()
        spark_version_info = VersionInfo(int(major), int(minor), int(patch), release, build="0")
    except (RuntimeError, AttributeError):
        spark_version_info = VersionInfo(major=3, minor=0, patch=1, release="unknown", build="0")
        logger = logging.getLogger(__name__)
        logger.warning("Could not parse spark version - using assumed Spark Version : %s", spark_version_info)

    return spark_version_info



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This module defines the package contents for the test data generator library distributions package
"""


from .normal_distribution import Normal
from .data_distribution import DataDistribution
from .beta import Beta
from .gamma import Gamma
from .exponential_distribution import Exponential


__all__ = ["normal_distribution",
           "gamma",
           "beta",
           "data_distribution",
           "exponential_distribution"]



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the base class for statistical distributions

Each inherited version of the DataDistribution object is used to generate random numbers drawn from
a specific distribution.

As the test data generator needs to scale the set of values generated across different data ranges,
the generate function is intended to generate values scaled to values between 0 and 1.

AS some distributions don't have easily predicted bounds, we scale the random data sets
by taking the minimum and maximum value of each generated data set and using that as the range for the generated data.

For some distributions, there may be alternative more efficient mechanisms for scaling the data to the [0, 1] interval.

Some data distributions are scaled to the [0,1] interval as part of their data generation
and no further scaling is needed.
"""
import copy
import pyspark.sql.functions as F
import numpy as np


class DataDistribution(object):
    """ Base class for all distributions"""
    def __init__(self):
        self._rounding = False
        self._randomSeed = None

    @staticmethod
    def get_np_random_generator(random_seed):
        """ Get numpy random number generator

        :param random_seed: Numeric random seed to use. If < 0, then no random
        :return:
        """
        assert random_seed is None or type(random_seed) in [ np.int32, np.int64, int],\
               f"`randomSeed` must be int or int-like not {type(random_seed)}"
        from numpy.random import default_rng
        if random_seed not in (-1, -1.0):
            rng = default_rng(random_seed)
        else:
            rng = default_rng()

        return rng

    def generateNormalizedDistributionSample(self):
        """ Generate sample of data for distribution

        :return: random samples from distribution scaled to values between 0 and 1
        """
        if self.randomSeed == -1 or self.randomSeed is None:
            newDef = F.expr("rand()")
        else:
            assert type(self.randomSeed) in [int, float],  "random seed should be numeric"
            newDef = F.expr(f"rand({self.randomSeed})")
        return newDef

    def withRounding(self, rounding):
        """ Create copy of object and set the rounding attribute

        :param rounding: rounding value to set. Should be True or False
        :return: new instance of data distribution object with rounding set
        """
        new_distribution_instance = copy.copy(self)
        new_distribution_instance._rounding = rounding
        return new_distribution_instance

    @property
    def rounding(self):
        """get the `rounding` attribute """
        return self._rounding

    def withRandomSeed(self, seed):
        """ Create copy of object and set the random seed attribute

        :param seed: random generator seed value to set. Should be integer,  float or None
        :return: new instance of data distribution object with rounding set
        """
        new_distribution_instance = copy.copy(self)
        new_distribution_instance._randomSeed = seed
        return new_distribution_instance

    @property
    def randomSeed(self):
        """get the `randomSeed` attribute """
        return self._randomSeed



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the `SparkSingleton` class

This is primarily meant for situations where the test data generator is run on a standalone environment
for use cases like unit testing rather than in a Databricks workspace environment
"""

import os
import logging
from pyspark.sql import SparkSession


class SparkSingleton:
    """A singleton class which returns one Spark session instance"""

    @classmethod
    def getInstance(cls):
        """Create a Spark instance for Datalib.

        :returns: A Spark instance
        """

        return SparkSession.builder.getOrCreate()

    @classmethod
    def getLocalInstance(cls, appName="new Spark session", useAllCores=True):
        """Create a machine local Spark instance for Datalib.
        By default, it uses `n-1` cores  of the available cores for the spark session,
        where `n` is total cores available.

        :param useAllCores:  If `useAllCores` is True, then use all cores rather than `n-1` cores
        :returns: A Spark instance
        """
        cpu_count = os.cpu_count()

        if useAllCores:
            spark_core_count = cpu_count
        else:
            spark_core_count = cpu_count - 1

        logger = logging.getLogger(__name__)
        logger.info("Spark core count: %d", spark_core_count)

        sparkSession = SparkSession.builder \
            .master(f"local[{spark_core_count}]") \
            .appName(appName) \
            .config("spark.sql.warehouse.dir", "/tmp/spark-warehouse") \
            .getOrCreate()

        return sparkSession



#
# Copyright (C) 2019 Databricks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This module defines library constants

"""

# default random seed
DEFAULT_RANDOM_SEED = 42
RANDOM_SEED_RANDOM = -1
RANDOM_SEED_RANDOM_FLOAT = -1.0
RANDOM_SEED_FIXED = "fixed"
RANDOM_SEED_HASH_FIELD_NAME = "hash_fieldname"

# constants related to seed column
DEFAULT_SEED_COLUMN = "id"

# this is the column name produced by `spark.range`
# dont change unless semantics of `spark.range` changes
SPARK_RANGE_COLUMN = "id"

# minimum versions for version checks
MIN_PYTHON_VERSION = (3, 8)
MIN_SPARK_VERSION = (3, 1, 2)

# options for randon data generation
OPTION_RANDOM = "random"
OPTION_RANDOM_SEED_METHOD = "randomSeedMethod"
OPTION_RANDOM_SEED = "randomSeed"

INFER_DATATYPE = "__infer__"

# default parallelism when sparkContext is not available
SPARK_DEFAULT_PARALLELISM = 200



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the `SchemaParser` class
"""

import re
import pyparsing as pp
from pyspark.sql.types import LongType, FloatType, IntegerType, StringType, DoubleType, BooleanType, ShortType, \
    TimestampType, DateType, DecimalType, ByteType, BinaryType, StructField, StructType, MapType, ArrayType


class SchemaParser(object):
    """ SchemaParser class

        Creates pyspark SQL datatype from string
    """
    _type_parser = None

    @classmethod
    def getTypeDefinitionParser(cls):
        """
        Define a pyparsing based parser for Spark SQL type definitions

        Allowable constructs for generated type parser are:
         * `string`, `varchar`, `char`, `nvarchar`,
         * `int`, `integer`,
         * `bigint`, `long`,
         * `bool`, `boolean`,
         * `smallint`, `short`
         * `binary`
         * `tinyint`, `byte`
         * `date`
         * `timestamp`, `datetime`,
         * `double`, `float`, `date`, `short`, `byte`,
         * `decimal` or `decimal(p)` or `decimal(p, s)` or `number(p, s)`
         * `map<type1, type2>` where type1 and type2 are type definitions of form accepted by parser
         * `array<type1>` where type1 is type definitions of form accepted by parser
         * `struct<a:binary, b:int, c:float>`

         Type definitions may be nested recursively - for example, the following are valid type definitions:
         * `array<array<int>>
         * `struct<a:array<int>, b:int, c:float>`
         * `map<string, struct<a:array<int>, b:int, c:float>>`

        :return: parser

        See the package pyparsing for details of how the parser mechanism works
        `https://pypi.org/project/pyparsing/`
        """
        # define grammar for parsing type definitions of the form "INT", "decimal(10,4)"
        # and "map<string, string>" etc
        if cls._type_parser is None:
            int_keyword = pp.CaselessKeyword("int") ^ pp.CaselessKeyword("integer")
            int_keyword.setParseAction(lambda s: "int")

            bigint_keyword = pp.CaselessKeyword("bigint") ^ pp.CaselessKeyword("long")
            bigint_keyword.setParseAction(lambda s: "bigint")

            binary_keyword = pp.CaselessKeyword("binary")

            boolean_keyword = pp.CaselessKeyword("boolean") ^ pp.CaselessKeyword("bool")
            boolean_keyword.setParseAction(lambda s: "boolean")

            date_keyword = pp.CaselessKeyword("date")
            double_keyword = pp.CaselessKeyword("double")

            float_keyword = pp.CaselessKeyword("float") ^ pp.CaselessKeyword("real")
            float_keyword.setParseAction(lambda s: "float")

            smallint_keyword = pp.CaselessKeyword("smallint") ^ pp.CaselessKeyword("short")
            smallint_keyword.setParseAction(lambda s: "smallint")

            timestamp_keyword = pp.CaselessKeyword("timestamp") ^ pp.CaselessKeyword("datetime")
            timestamp_keyword.setParseAction(lambda s: "timestamp")

            tinyint_keyword = pp.CaselessKeyword("tinyint") ^ pp.CaselessKeyword("byte")
            tinyint_keyword.setParseAction(lambda s: "tinyint")

            lbracket = (pp.Literal("(").suppress()).setName("lPar")
            rbracket = pp.Literal(")").suppress().setName("rPar")
            comma = pp.Literal(",").suppress().setName("comma")
            number = pp.Word(pp.nums).setName("number_literal")

            # handle string types of the form "STRING", "char(10)" etc
            string_keyword = pp.oneOf(["string", "nvarchar", "varchar", "char"], caseless=True, asKeyword=True)
            string_keyword.setParseAction(lambda s: "string")
            string_type_expr = string_keyword + pp.Optional(lbracket + number + rbracket)

            # handle decimal types of the form "decimal(10)", "real", "numeric(10,3)"
            decimal_keyword = pp.MatchFirst(
                [pp.CaselessKeyword("decimal"), pp.CaselessKeyword("dec"), pp.CaselessKeyword("number"),
                 pp.CaselessKeyword("numeric")])
            decimal_keyword.setParseAction(lambda s: "decimal")
            # first number is precision , default 10; second number is scale, default 0
            decimal_type_expr = decimal_keyword + pp.Optional(
                lbracket + number + pp.Optional(comma + number, "0") + rbracket)

            primitive_type_keyword = (int_keyword ^ bigint_keyword ^ binary_keyword ^ boolean_keyword
                                      ^ date_keyword ^ float_keyword ^ smallint_keyword ^ timestamp_keyword
                                      ^ tinyint_keyword ^ string_type_expr ^ decimal_type_expr ^ double_keyword
                                      ).setName("primitive_type_defn")

            # handle more complex type definitions such as struct, map and array

            # declare forward reference to type expression
            type_expr = pp.Forward()

            l_angle = pp.Literal("<").suppress()
            r_angle = pp.Literal(">").suppress()

            # handle arrays
            array_keyword = pp.CaselessKeyword("array")
            array_expr = array_keyword + pp.Group(l_angle + type_expr + r_angle)
            array_expr.setName("array_type_expr")

            # handle maps
            map_keyword = pp.CaselessKeyword("map")
            map_expr = map_keyword + l_angle + pp.Group(type_expr) + comma + pp.Group(type_expr) + r_angle
            map_expr.setName("map_type_expr")

            # handle SQL identifiers such as "a14" and "`a test`"
            ident = pp.Word(pp.alphas, pp.alphanums + "_") | pp.QuotedString(quoteChar="`", escQuote="``")

            colon = pp.Literal(":").suppress()

            # handle structs
            struct_keyword = pp.CaselessKeyword("struct")
            struct_expr = struct_keyword + l_angle + pp.Group(
                pp.delimitedList(pp.Group(ident + pp.Optional(colon) + pp.Group(type_expr)))) + r_angle

            # try to capture invalid type name for better error reporting
            invalid_type = pp.Word(pp.alphas, pp.alphanums + "_", asKeyword=True)

            # use left recursion to handle nesting of types
            type_expr <<= pp.MatchFirst([primitive_type_keyword, array_expr, map_expr, struct_expr, invalid_type])

            type_parser = pp.StringStart() + type_expr + pp.StringEnd()
            cls._type_parser = type_parser
        return cls._type_parser

    @classmethod
    def _parse_ast(cls, ast):
        """
        Parse Abstract Syntax Tree generated from earlier parse
        :param ast: AST generated by parsing type definition
        :return: Pyspark Type Definition - e.g StringType() etc
        """
        assert len(ast) > 0, "AST tree must contain elements"

        first_token = ast[0]

        if first_token == "string":
            retval = StringType()
        elif first_token == "int":
            retval = IntegerType()
        elif first_token == "bigint":
            retval = LongType()
        elif first_token == "boolean":
            retval = BooleanType()
        elif first_token == "binary":
            retval = BinaryType()
        elif first_token == "timestamp":
            retval = TimestampType()
        elif first_token == "decimal":
            if len(ast) == 1:
                retval = DecimalType(10, 0)
            # note if `decimal(10)` is passed, parser will expand to `decimal(10,0)`
            # so there is no need to handle case here len(ast) == 2
            elif len(ast) == 3:
                retval = DecimalType(int(ast[1]), int(ast[2]))
            else:
                raise ValueError(f"Cannot parse decimal type {list(ast)}")
        elif first_token == "double":
            retval = DoubleType()
        elif first_token == "float":
            retval = FloatType()
        elif first_token == "date":
            retval = DateType()
        elif first_token == "smallint":
            retval = ShortType()
        elif first_token == "tinyint":
            retval = ByteType()
        elif first_token == "interval":
            raise ValueError("Interval is invalid type for field definition")
        elif first_token == "array":
            try:
                assert len(ast) == 2, "array must have nested type"
                inner_type = cls._parse_ast(ast[1])
                retval = ArrayType(inner_type)
            except Exception as e:
                raise ValueError(f"Could not construct array type definition due to `{e}`") from e
        elif first_token == "map":
            try:
                assert len(ast) == 3, "map must have 2 inner types"
                inner_type1 = cls._parse_ast(ast[1])
                inner_type2 = cls._parse_ast(ast[2])
                retval = MapType(inner_type1, inner_type2)
            except Exception as e:
                raise ValueError(f"Could not construct map type definition due to `{e}`") from e
        elif first_token == "struct":
            try:
                assert len(ast) == 2, f"struct must have nested list of fields : {len(ast)}"
                field_list_defns = ast[1]
                fields = []
                assert len(field_list_defns) > 0, "field list must be non-empty"

                for field_defn in field_list_defns:
                    field_name = field_defn[0]
                    field_type = cls._parse_ast(field_defn[1])
                    fields.append(StructField(field_name, field_type))
                retval = StructType(fields)
            except Exception as e:
                raise ValueError(f"Could not construct struct type definition due to `{e}`") from e
        else:
            raise ValueError(f" Invalid or unsupported type `{first_token}` found while processing `{list(ast)}`")
        return retval

    @classmethod
    def columnTypeFromString(cls, type_string):
        """ Generate a Spark SQL data type from a string

        Allowable options for `type_string` parameter are:
         * `string`, `varchar`, `char`, `nvarchar`,
         * `int`, `integer`,
         * `bigint`, `long`,
         * `bool`, `boolean`,
         * `smallint`, `short`
         * `binary`
         * `tinyint`, `byte`
         * `date`
         * `timestamp`, `datetime`,
         * `double`, `float`, `date`, `short`, `byte`,
         * `decimal` or `decimal(p)` or `decimal(p, s)` or `number(p, s)`
         * `map<type1, type2>` where type1 and type2 are type definitions of form accepted by parser
         * `array<type1>` where type1 is type definitions of form accepted by parser
         * `struct<a:binary, b:int, c:float>`

         Type definitions may be nested recursively - for example, the following are valid type definitions:
         * `array<array<int>>
         * `struct<a:array<int>, b:int, c:float>`
         * `map<string, struct<a:array<int>, b:int, c:float>>`

        :param type_string: String representation of SQL type such as 'integer' etc.
        :returns: Spark SQL type
        """
        assert type_string is not None, "`type_string` must be specified"

        parser = cls.getTypeDefinitionParser()

        # generate abstract syntax tree (AST) from parsed definition
        try:
            ast = parser.parseString(type_string)
        except Exception as e:
            raise ValueError(f"Invalid type definition `{type_string}`", e) from e

        type_construct = cls._parse_ast(ast)

        return type_construct

    @classmethod
    def _cleanseSQL(cls, sql_string):
        """ Cleanse sql string removing string literals so that they are not considered as part of potential column
            references
        :param sql_string: String representation of SQL expression
        :returns: cleansed string

        Any strings identified are replaced with `' '`
        """
        assert sql_string is not None, "`sql_string` must be specified"

        # skip over quoted identifiers even if they contain quotes
        quoted_ident = pp.QuotedString(quoteChar="`", escQuote="``")
        quoted_ident.setParseAction(lambda s, loc, toks: f"`{toks[0]}`")

        stringForm1 = pp.Literal('r') + pp.QuotedString(quoteChar="'")
        stringForm2 = pp.Literal('r') + pp.QuotedString(quoteChar='"')
        stringForm3 = pp.QuotedString(quoteChar="'", escQuote=r"\'")
        stringForm4 = pp.QuotedString(quoteChar='"', escQuote=r'\"')
        stringForm = stringForm1 ^ stringForm2 ^ stringForm3 ^ stringForm4
        stringForm.setParseAction(lambda s, loc, toks: "' '")

        parser = quoted_ident ^ stringForm

        transformed_string = parser.transformString(sql_string)

        return transformed_string

    @classmethod
    def columnsReferencesFromSQLString(cls, sql_string, filterItems=None):
        """ Generate a list of possible column references from a SQL string

        This method finds all condidate references to SQL columnn ids in the string

        To avoid the overhead of a full SQL parser, the implementation will simply look for possible field names

        Further improvements may eliminate some common syntax but in current form, reserved words will
        also be returned as possible column references.

        So any uses of this must not assume that all possible references are valid column references

        :param sql_string: String representation of SQL expression
        :param filterItems: filter results to only results in items listed
        :returns: list of possible column references
        """
        assert sql_string is not None, "`sql_string` must be specified"
        assert filterItems is None or isinstance(filterItems, (list, set))

        cleansed_sql_string = cls._cleanseSQL(sql_string)

        ident = pp.Word(pp.alphas, pp.alphanums + "_") | pp.QuotedString(quoteChar="`", escQuote="``")
        parser = ident

        references = parser.searchString(cleansed_sql_string)

        results = set([item for sublist in references for item in sublist])

        if filterItems is not None:
            filtered_results = results.intersection(set(filterItems))
            return list(filtered_results)
        else:
            return list(results)

    @classmethod
    def parseCreateTable(cls, sparkSession, source_schema):
        """ Parse a schema from a schema string

            :param sparkSession: spark session to use
            :param source_schema: should be a table definition minus the create table statement
            :returns: Spark SQL schema instance
        """
        assert (source_schema is not None), "`source_schema` must be specified"
        assert (sparkSession is not None), "`sparkSession` must be specified"
        lines = [x.strip() for x in source_schema.split("\n") if x is not None]
        table_defn = " ".join(lines)

        # get table name from s
        table_name_match = re.search(r"^\s*create\s*(temporary\s*)?table\s*([a-zA-Z0-9_]*)\s*(\(.*)$", table_defn,
                                     flags=re.IGNORECASE)

        if table_name_match:
            table_name = table_name_match.group(2)
            table_body = table_name_match.group(3)
        else:
            raise ValueError("Table name could not be found")

        sparkSession.sql(f"create temporary table {table_name}{table_body}using csv")

        result = sparkSession.sql(f"select * from {table_name}")
        result.printSchema()
        result_schema = result.schema

        sparkSession.sql(f"drop table {table_name}")

        return result_schema



# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file defines the `DataGenError` and `DataGenerator` classes
"""
import copy
import logging
import re

from pyspark.sql.types import LongType, IntegerType, StringType, StructType, StructField, DataType

from ._version import _get_spark_version
from .column_generation_spec import ColumnGenerationSpec
from .datagen_constants import DEFAULT_RANDOM_SEED, RANDOM_SEED_FIXED, RANDOM_SEED_HASH_FIELD_NAME, \
    DEFAULT_SEED_COLUMN, SPARK_RANGE_COLUMN, MIN_SPARK_VERSION, \
    OPTION_RANDOM, OPTION_RANDOM_SEED, OPTION_RANDOM_SEED_METHOD, \
    INFER_DATATYPE, SPARK_DEFAULT_PARALLELISM
from .html_utils import HtmlUtils
from .schema_parser import SchemaParser
from .spark_singleton import SparkSingleton
from .utils import ensure, topologicalSort, DataGenError, deprecated, split_list_matching_condition

_OLD_MIN_OPTION = 'min'
_OLD_MAX_OPTION = 'max'

_STREAMING_TIMESTAMP_COLUMN = "_source_timestamp"


class DataGenerator:
    """ Main Class for test data set generation

    This class acts as the entry point to all test data generation activities.

    :param sparkSession: spark Session object to use
    :param name: is name of data set
    :param randomSeedMethod: = seed method for random numbers - either None, 'fixed', 'hash_fieldname'
    :param rows: = amount of rows to generate
    :param startingId: = starting value for generated seed column
    :param randomSeed: = seed for random number generator
    :param partitions: = number of partitions to generate, if not provided, uses `spark.sparkContext.defaultParallelism`
    :param verbose: = if `True`, generate verbose output
    :param batchSize: = UDF batch number of rows to pass via Apache Arrow to Pandas UDFs
    :param debug: = if set to True, output debug level of information
    :param seedColumnName: = if set, this should be the name of the `seed` or logical `id` column. Defaults to `id`
    :param random: = if set, specifies default value of `random` attribute for all columns where not set

    By default the seed column is named `id`. If you need to use this column name in your generated data,
    it is recommended that you use a different name for the seed column - for example `_id`.

    This may be specified by setting the `seedColumnName` attribute to `_id`

    Note: in a shared spark session, the sparkContext is not available, so the default parallelism is set to 200.
    We recommend passing an explicit value for `partitions` in this case.
    """

    # class vars
    _nextNameIndex = 0
    _untitledNamePrefix = "Untitled"
    _randomSeed = DEFAULT_RANDOM_SEED

    _allowed_keys = ["startingId", "rowCount", "output_id"]

    # set up logging

    # restrict spurious messages from java gateway
    logging.getLogger("py4j").setLevel(logging.WARNING)

    # logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.NOTSET)

    def __init__(self, sparkSession=None, name=None, randomSeedMethod=None,
                 rows=1000000, startingId=0, randomSeed=None, partitions=None, verbose=False,
                 batchSize=None, debug=False, seedColumnName=DEFAULT_SEED_COLUMN,
                 random=False,
                 **kwargs):
        """ Constructor for data generator object """

        # set up logging
        self.verbose = verbose
        self.debug = debug

        self._setupLogger()
        self._seedColumnName = seedColumnName
        self._outputStreamingFields = False

        if seedColumnName != DEFAULT_SEED_COLUMN:
            self.logger.info(f"Using '{self._seedColumnName}' for seed column in place of '{DEFAULT_SEED_COLUMN}")

        self.name = name if name is not None else self.generateName()
        self._rowCount = rows
        self.starting_id = startingId
        self.__schema__ = None

        if sparkSession is None:
            sparkSession = SparkSingleton.getLocalInstance()

        self.sparkSession = sparkSession

        # if the active Spark session is stopped, you may end up with a valid SparkSession object but the underlying
        # SparkContext will be invalid
        assert sparkSession is not None, "Spark session not initialized"

        self.partitions = partitions if partitions is not None else self._getDefaultSparkParallelism(sparkSession)

        # check for old versions of args
        if "starting_id" in kwargs:
            self.logger.warning("starting_id is deprecated - use option 'startingId' instead")
            startingId = kwargs["starting_id"]

        if "seed" in kwargs:
            self.logger.warning("seed is deprecated - use option 'randomSeed' instead")
            randomSeed = kwargs["seed"]

        if "seed_method" in kwargs:
            self.logger.warning("seed_method is deprecated - use option 'seedMethod' instead")
            seedMethod = kwargs["seed_method"]

        if "batch_size" in kwargs:
            self.logger.warning("batch_size is deprecated - use option 'batchSize' instead")
            batchSize = kwargs["batch_size"]

        if "use_pandas" in kwargs or "usePandas" in kwargs:
            self.logger.warning("option 'usePandas' is deprecated - Pandas will always be used")

        if "generateWithSelects" in kwargs or "generateWithSelects" in kwargs:
            self.logger.warning("option 'generateWithSelects' switch is deprecated - selects will always be used")

        self._seedMethod = randomSeedMethod

        # set default random setting
        self._defaultRandom = random if random is not None else False

        if randomSeed is None:
            self._instanceRandomSeed = self._randomSeed

            if randomSeedMethod is None:
                self._seedMethod = RANDOM_SEED_HASH_FIELD_NAME
            else:
                self._seedMethod = randomSeedMethod
        else:
            self._instanceRandomSeed = randomSeed

            # if a valid random seed was supplied but no seed method was applied, make the seed method "fixed"
            if randomSeedMethod is None:
                self._seedMethod = "fixed"

        if self._seedMethod not in [None, RANDOM_SEED_FIXED, RANDOM_SEED_HASH_FIELD_NAME]:
            msg = f"seedMethod should be None, '{RANDOM_SEED_FIXED}' or '{RANDOM_SEED_HASH_FIELD_NAME}' "
            raise DataGenError(msg)

        self._columnSpecsByName = {}
        self._allColumnSpecs = []
        self._buildPlan = []
        self.executionHistory = []
        self._options = {}
        self._buildOrder = []
        self._inferredSchemaFields = []
        self.buildPlanComputed = False

        # lets add the seed column
        self.withColumn(self._seedColumnName, LongType(), nullable=False, implicit=True, omit=True, noWarn=True)
        self._batchSize = batchSize

        # set up spark session
        self._setupSparkSession(sparkSession)

        # set up use of pandas udfs
        self._setupPandas(batchSize)

    @property
    def seedColumnName(self):
        """ return the name of data generation seed column"""
        return self._seedColumnName

    @classmethod
    def _checkSparkVersion(cls, sparkVersion, minSparkVersion):
        """
        check spark version
        :param sparkVersion: spark version string
        :param minSparkVersion: min spark version as tuple
        :return: True if version passes minVersion

        Layout of version string must be compatible "xx.xx.xx.patch"
        """
        sparkVersionInfo = _get_spark_version(sparkVersion)

        if sparkVersionInfo < minSparkVersion:
            logging.warning(f"*** Minimum version of Python supported is {minSparkVersion} - found version %s ",
                            sparkVersionInfo)
            return False

        return True

    def _setupSparkSession(self, sparkSession):
        """
        Set up spark session
        :param sparkSession: spark session to use
        :return: nothing
        """
        if sparkSession is None:
            sparkSession = SparkSingleton.getInstance()

        assert sparkSession is not None, "Spark session not initialized"

        self.sparkSession = sparkSession

        # check if the spark version meets the minimum requirements and warn if not
        sparkVersion = sparkSession.version
        self._checkSparkVersion(sparkVersion, MIN_SPARK_VERSION)

    def _setupPandas(self, pandasBatchSize):
        """
        Set up pandas
        :param pandasBatchSize: batch size for pandas, may be None
        :return: nothing
        """
        assert pandasBatchSize is None or type(pandasBatchSize) is int, \
            "If pandas_batch_size is specified, it must be an integer"
        self.logger.info("*** using pandas udf for custom functions ***")
        self.logger.info("Spark version: %s", self.sparkSession.version)
        if str(self.sparkSession.version).startswith("3"):
            self.logger.info("Using spark 3.x")
            self.sparkSession.conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")
        else:
            self.sparkSession.conf.set("spark.sql.execution.arrow.enabled", "true")

        if self._batchSize is not None:
            self.sparkSession.conf.set("spark.sql.execution.arrow.maxRecordsPerBatch", self._batchSize)

    def _setupLogger(self):
        """Set up logging

        This will set the logger at warning, info or debug levels depending on the instance construction parameters
        """
        self.logger = logging.getLogger("DataGenerator")
        if self.debug:
            self.logger.setLevel(logging.DEBUG)
        elif self.verbose:
            self.logger.setLevel(logging.INFO)
        else:
            self.logger.setLevel(logging.WARNING)

    @staticmethod
    def _getDefaultSparkParallelism(sparkSession):
        """Get the default parallelism for a spark session, if spark session supports getting the sparkContext
        :param sparkSession: spark session
        :return: default parallelism
        """
        try:
            if sparkSession.sparkContext is not None:
                return sparkSession.sparkContext.defaultParallelism
            else:
                return SPARK_DEFAULT_PARALLELISM
        except Exception as err:  # pylint: disable=broad-exception-caught
            err_msg = f"Error getting default parallelism, using default setting of {SPARK_DEFAULT_PARALLELISM}"
            logging.warning(err_msg)
            return SPARK_DEFAULT_PARALLELISM

    @classmethod
    def useSeed(cls, seedVal):
        """ set seed for random number generation

            Arguments:
            :param seedVal: - new value for the random number seed
        """
        cls._randomSeed = seedVal

    @deprecated("Use `useSeed` instead")
    @classmethod
    def use_seed(cls, seedVal):
        """ set seed for random number generation

            Arguments:
            :param seedVal: - new value for the random number seed

        """
        cls._randomSeed = seedVal

    @classmethod
    def reset(cls):
        """ reset any state associated with the data """
        cls._nextNameIndex = 0

    @classmethod
    def generateName(cls):
        """ get a name for the data set

            Uses the untitled name prefix and nextNameIndex to generate a dummy dataset name

            :returns: string containing generated name
        """
        cls._nextNameIndex += 1
        new_name = cls._untitledNamePrefix + '_' + str(cls._nextNameIndex)
        return new_name

    # noinspection PyAttributeOutsideInit
    def clone(self):
        """Make a clone of the data spec via deep copy preserving same spark session

        :returns: deep copy of test data generator definition
        """
        old_spark_session = self.sparkSession
        old_logger = self.logger
        new_copy = None
        try:
            # temporarily set the spark session to null
            self.sparkSession = None

            # set logger to None before copy, disable pylint warning to ensure not triggered for this statement
            self.logger = None  # pylint: disable=attribute-defined-outside-init
            new_copy = copy.deepcopy(self)
            new_copy.sparkSession = old_spark_session
            new_copy.buildPlanComputed = False
            new_copy._setupLogger()
        finally:
            # now set it back
            self.sparkSession = old_spark_session
            # set logger to old value, disable pylint warning to ensure not triggered for this statement
            self.logger = old_logger  # pylint: disable=attribute-defined-outside-init
        return new_copy

    @property
    def randomSeed(self):
        """ return the data generation spec random seed"""
        return self._instanceRandomSeed

    @property
    def random(self):
        """ return the data generation spec default random setting for columns to be used
            when an explicit `random` attribute setting is not supplied
        """
        return self._defaultRandom

    def _markForPlanRegen(self):
        """Mark that build plan needs to be regenerated

        :returns: modified in-place instance of test data generator allowing for chaining of calls following
                  Builder pattern
        """
        self.buildPlanComputed = False
        return self

    def explain(self, suppressOutput=False):
        """Explain the test data generation process

        :param suppressOutput: If True, suppress display of build plan
        :returns: String containing explanation of test data generation for this specification
        """
        if not self.buildPlanComputed:
            self.computeBuildPlan()

        rc = self._rowCount
        tasks = self.partitions
        output = ["",
                  "Data generation plan", "====================",
                  f"spec=DateGenerator(name={self.name}, rows={rc}, startingId={self.starting_id}, partitions={tasks})",
                  ")",
                  "",
                  f"seed column: {self._seedColumnName}", "",
                  f"column build order: {self._buildOrder}", "", "build plan:"]

        for plan_action in self._buildPlan:
            output.append(" ==> " + plan_action)
        output.extend(["", "execution history:", ""])
        for build_action in self.executionHistory:
            output.append(" ==> " + build_action)
        output.append("")
        output.append("====================")
        output.append("")

        explain_results = "\n".join(output)
        if not suppressOutput:
            print(explain_results)

        return explain_results

    def withRowCount(self, rc):
        """Modify the row count - useful when starting a new spec from a clone

        :param rc: The count of rows to generate
        :returns: modified in-place instance of test data generator allowing for chaining of calls following
                  Builder pattern
        """
        self._rowCount = rc
        return self

    @deprecated('Use `withRowCount` instead')
    def setRowCount(self, rc):
        """Modify the row count - useful when starting a new spec from a clone

        .. warning::
           Method is deprecated - use `withRowCount` instead

        :param rc: The count of rows to generate
        :returns: modified in-place instance of test data generator allowing for chaining of calls following
                  Builder pattern

        """
        self.logger.warning("method `setRowCount` is deprecated, use `withRowCount` instead")
        return self.withRowCount(rc)

    @property
    def rowCount(self):
        """ Return the row count

        This may differ from the original specified row counts, if counts need to be adjusted for purposes of
        keeping the ratio of rows to unique keys correct or other heuristics
        """
        return self._rowCount

    def withIdOutput(self):
        """ output seed column field (defaults to `id`) as a column in the generated data set if specified

        If this is not called, the seed column field is omitted from the final generated data set

        :returns: modified in-place instance of test data generator allowing for chaining of calls
                  following Builder pattern
        """
        self._columnSpecsByName[self._seedColumnName].omit = False
        self._markForPlanRegen()

        return self

    def option(self, optionKey, optionValue):
        """ set option to option value for later processing

        :param optionKey: key for option
        :param optionValue: value for option
        :returns: modified in-place instance of test data generator allowing for chaining of calls
                  following Builder pattern
        """
        ensure(optionKey in self._allowed_keys)
        self._options[optionKey] = optionValue
        self._markForPlanRegen()
        return self

    def options(self, **kwargs):
        """ set options in bulk

        Allows for multiple options with option=optionValue style of option passing

        :returns: modified in-place instance of test data generator allowing for chaining of calls
                  following Builder pattern
        """
        for key, value in kwargs.items():
            self.option(key, value)
        self._markForPlanRegen()
        return self

    def _processOptions(self):
        """ process options to give effect to the options supplied earlier

        :returns: modified in-place instance of test data generator allowing for chaining of calls
                  following Builder pattern
        """
        self.logger.info("options: %s", str(self._options))

        for key, value in self._options.items():
            if key == "startingId":
                self.starting_id = value
            elif key in ["rowCount", "row_count"]:
                self._rowCount = value
        return self

    def describe(self):
        """ return description of the dataset generation spec

        :returns: Dict object containing key attributes of test data generator instance
        """

        return {
            'name': self.name,
            'rowCount': self._rowCount,
            'schema': self.schema,
            'randomSeed': self._instanceRandomSeed,
            'partitions': self.partitions,
            'columnDefinitions': self._columnSpecsByName,
            'debug': self.debug,
            'verbose': self.verbose
        }

    def __repr__(self):
        """ return the repr string for the class"""
        name = getattr(self, "name", "None")
        rows = getattr(self, "_rowCount", "None")
        partitions = getattr(self, "partitions", "None")
        return f"DataGenerator(name='{name}', rows={rows}, partitions={partitions})"

    def _checkFieldList(self):
        """Check field list for common errors

        Does not return anything but will assert / raise exceptions if errors occur
        """
        ensure(self._inferredSchemaFields is not None, "schemaFields should be non-empty")
        ensure(type(self._inferredSchemaFields) is list, "schemaFields should be list")

    @property
    def schemaFields(self):
        """ get list of schema fields for final output schema

        :returns: list of fields in schema
        """
        self._checkFieldList()
        return [fd for fd in self._inferredSchemaFields if not self._columnSpecsByName[fd.name].isFieldOmitted]

    @property
    def schema(self):
        """ infer spark output schema definition from the field specifications

        :returns: Spark SQL `StructType` for schema

        ..note::
          If the data generation specification contains columns for which the datatype is inferred, the schema type
          for inferred columns may not be correct until the build command has completed.

        """
        return StructType(self.schemaFields)

    @property
    def inferredSchema(self):
        """ infer spark interim schema definition from the field specifications

        ..note::
          If the data generation specification contains columns for which the datatype is inferred, the schema type
          for inferred columns may not be correct until the build command has completed.

        """
        self._checkFieldList()
        return StructType(self._inferredSchemaFields)

    def __getitem__(self, key):
        """ implement the built in derefernce by key behavior """
        ensure(key is not None, "key should be non-empty")
        return self._columnSpecsByName[key]

    def getColumnType(self, colName):
        """ Get column Spark SQL datatype for specified column

        :param colName: name of column as string
        :returns: Spark SQL datatype for named column
        """
        ct = self._columnSpecsByName[colName].datatype
        return ct if ct is not None else IntegerType()

    def isFieldExplicitlyDefined(self, colName):
        """ return True if column generation spec has been explicitly defined for column, else false

        .. note::
           A column is not considered explicitly defined if it was inferred from a schema or added
           with a wildcard statement. This impacts whether the column can be redefined.
        """
        ensure(colName is not None, "colName should be non-empty")
        col_def = self._columnSpecsByName.get(colName, None)
        return not col_def.implicit if col_def is not None else False

    def getInferredColumnNames(self):
        """ get list of output columns """
        return [fd.name for fd in self._inferredSchemaFields]

    @staticmethod
    def flatten(lst):
        """ flatten list

        :param lst: list to flatten
        """
        return [item for sublist in lst for item in sublist]

    def getColumnSpec(self, name):
        """ get column spec for column having name supplied

        :param name: name of column to find spec for
        :return: column spec for named column if any
        """
        assert name is not None and len(name.strip()) > 0, "column name must be non empty string"
        return self._columnSpecsByName[name]

    def getOutputColumnNames(self):
        """ get list of output columns by flattening list of lists of column names
            normal columns will have a single column name but column definitions that result in
            multiple columns will produce a list of multiple names

            :returns: list of column names to be output in generated data set
        """
        return self.flatten([self._columnSpecsByName[fd.name].getNames()
                             for fd in
                             self._inferredSchemaFields if not self._columnSpecsByName[fd.name].isFieldOmitted])

    def getOutputColumnNamesAndTypes(self):
        """ get list of output columns by flattening list of lists of column names and types
            normal columns will have a single column name but column definitions that result in
            multiple columns will produce a list of multiple names
        """
        return self.flatten([self._columnSpecsByName[fd.name].getNamesAndTypes()
                             for fd in
                             self._inferredSchemaFields if not self._columnSpecsByName[fd.name].isFieldOmitted])

    def withSchema(self, sch):
        """ populate column definitions and specifications for each of the columns in the schema

        :param sch: Spark SQL schema, from which fields are added
        :returns: modified in-place instance of test data generator allowing for chaining of calls
                  following Builder pattern
        """
        ensure(sch is not None, "schema sch should be non-empty")
        self.__schema__ = sch

        for fs in sch.fields:
            self.withColumn(fs.name, fs.dataType, implicit=True, omit=False, nullable=fs.nullable)
        return self

    def _computeRange(self, dataRange, minValue, maxValue, step):
        """ Compute merged range based on parameters

        :returns: effective minValue, maxValue, step as tuple
        """
        # TODO: may also need to check for instance of DataRange
        if dataRange is not None and isinstance(dataRange, range):
            if maxValue is not None or minValue != 0 or step != 1:
                raise ValueError("You cant specify both a range and minValue, maxValue or step values")

            return dataRange.start, dataRange.stop, dataRange.step
        else:
            return minValue, maxValue, step

    def withColumnSpecs(self, patterns=None, fields=None, matchTypes=None, **kwargs):
        """Add column specs for columns matching
           a) list of field names,
           b) one or more regex patterns
           c) type (as in pyspark.sql.types)

        :param patterns: patterns may specified a single pattern as a string or a list of patterns
                         that match the column names. May be omitted.
        :param fields: a string specifying an explicit field to match , or a list of strings specifying
                       explicit fields to match. May be omitted.
        :param matchTypes: a single Spark SQL datatype or list of Spark SQL data types to match. May be omitted.
        :returns: modified in-place instance of test data generator allowing for chaining of calls following
                  Builder pattern

        .. note::
           matchTypes may also take SQL type strings or a list of SQL type strings such as "array<integer>". However,
           you may not use ``INFER_DATYTYPE`` as part of the matchTypes list.

        You may also add a variety of options to further control the test data generation process.
        For full list of options, see :doc:`/reference/api/dbldatagen.column_spec_options`.

        """
        if fields is not None and type(fields) is str:
            fields = [fields]

        if OPTION_RANDOM not in kwargs:
            kwargs[OPTION_RANDOM] = self._defaultRandom

        # add support for deprecated legacy names
        if "match_types" in kwargs:
            assert matchTypes is None, "Argument 'match_types' is deprecated, use 'matchTypes' instead"
            matchTypes = kwargs["match_types"]
            del kwargs["match_types"]  # remove the legacy option from keyword args as they will be used later

        if matchTypes is not None and type(matchTypes) is not list:
            matchTypes = [matchTypes]  # if only one match type, make a list of that match type

        if patterns is not None and type(patterns) is str:
            patterns = ["^" + patterns + "$"]
        elif type(patterns) is list:
            patterns = ["^" + pat + "$" for pat in patterns]

        all_fields = self.getInferredColumnNames()
        effective_fields = [x for x in all_fields if
                            (fields is None or x in fields) and x != self._seedColumnName]

        if patterns is not None:
            effective_fields = [x for x in effective_fields for y in patterns if re.search(y, x) is not None]

        if matchTypes is not None:
            effective_types = []

            for typ in matchTypes:
                if isinstance(typ, str):
                    if typ == INFER_DATATYPE:
                        raise ValueError("You cannot use INFER_DATATYPE with the method `withColumnSpecs`")

                    effective_types.append(SchemaParser.columnTypeFromString(typ))
                else:
                    effective_types.append(typ)

            effective_fields = [x for x in effective_fields for y in effective_types
                                if self.getColumnType(x) == y]

        for f in effective_fields:
            self.withColumnSpec(f, implicit=True, **kwargs)
        return self

    def _checkColumnOrColumnList(self, columns, allowId=False):
        """ Check if column or columns refer to existing columns

        :param columns: a single column or list of columns as strings
        :param allowId: If True, allows the specialized seed column (which defaults to `id`) to be present in columns
        :returns: True if test passes
        """
        inferred_columns = self.getInferredColumnNames()
        if allowId and columns == self._seedColumnName:
            return True

        if type(columns) is list:
            for column in columns:
                ensure(column in inferred_columns,
                       f" column `{column}` must refer to defined column")
        else:
            ensure(columns in inferred_columns,
                   f" column `{columns}` must refer to defined column")
        return True

    def withColumnSpec(self, colName, minValue=None, maxValue=None, step=1, prefix=None,
                       random=None, distribution=None,
                       implicit=False, dataRange=None, omit=False, baseColumn=None, **kwargs):
        """ add a column specification for an existing column

        :returns: modified in-place instance of test data generator allowing for chaining of calls
                  following Builder pattern

        You may also add a variety of options to further control the test data generation process.
        For full list of options, see :doc:`/reference/api/dbldatagen.column_spec_options`.

        """
        ensure(colName is not None, "Must specify column name for column")
        ensure(colName in self.getInferredColumnNames(), f" column `{colName}` must refer to defined column")
        if baseColumn is not None:
            self._checkColumnOrColumnList(baseColumn)
        ensure(not self.isFieldExplicitlyDefined(colName), f"duplicate column spec for column `{colName}`")

        ensure(not isinstance(minValue, DataType),
               f"""unnecessary `datatype` argument specified for `withColumnSpec` for column `{colName}` -
                    Datatype parameter is only needed for `withColumn` and not permitted for `withColumnSpec`
               """)

        if random is None:
            random = self._defaultRandom

        # handle migration of old `min` and `max` options
        if _OLD_MIN_OPTION in kwargs:
            assert minValue is None, \
                "Only one of `minValue` and `minValue` can be specified. Use of `minValue` is preferred"
            minValue = kwargs[_OLD_MIN_OPTION]
            kwargs.pop(_OLD_MIN_OPTION, None)

        if _OLD_MAX_OPTION in kwargs:
            assert maxValue is None, \
                "Only one of `maxValue` and `maxValue` can be specified. Use of `maxValue` is preferred"
            maxValue = kwargs[_OLD_MAX_OPTION]
            kwargs.pop(_OLD_MAX_OPTION, None)

        new_props = {}
        new_props.update(kwargs)

        self.logger.info("adding column spec - `%s` with baseColumn : `%s`, implicit : %s , omit %s",
                         colName, baseColumn, implicit, omit)

        self._generateColumnDefinition(colName, self.getColumnType(colName), minValue=minValue, maxValue=maxValue,
                                       step=step, prefix=prefix,
                                       random=random, dataRange=dataRange,
                                       distribution=distribution, baseColumn=baseColumn,
                                       implicit=implicit, omit=omit, **new_props)
        return self

    def hasColumnSpec(self, colName):
        """returns true if there is a column spec for the column

        :param colName: name of column to check for
        :returns: True if column has spec, False otherwise
        """
        return colName in self._columnSpecsByName

    def withColumn(self, colName, colType=StringType(), minValue=None, maxValue=None, step=1,
                   dataRange=None, prefix=None, random=None, distribution=None,
                   baseColumn=None, nullable=True,
                   omit=False, implicit=False, noWarn=False,
                   **kwargs):
        """ add a new column to the synthetic data generation specification

        :param colName: Name of column to add. If this conflicts with the underlying seed column (`id`), it is
                        recommended that the seed column name is customized during the construction of the data
                        generator spec.
        :param colType: Data type for column. This may be specified as either a type from one of the possible
                        pyspark.sql.types (e.g. `StringType`, `DecimalType(10,3)` etc) or as a string containing a Spark
                        SQL type definition (i.e  `String`, `array<Integer>`, `map<String, Float>`)
        :param omit: if True, the column will be omitted from the final set of columns in the generated data.
                     Used to create columns that are used by other columns as intermediate results.
                     Defaults to False

        :param expr: Specifies SQL expression used to create column value. If specified, overrides the default rules
                     for creating column value. Defaults to None

        :param baseColumn: String or list of columns to control order of generation of columns. If not specified,
                           column is dependent on base seed column (which defaults to `id`)

        :returns: modified in-place instance of test data generator allowing for chaining of calls
                  following Builder pattern

        .. note::
           if the value ``None`` is used for the ``colType`` parameter, the method will try to use the underlying
           datatype derived from the base columns.

           If the value ``INFER_DATATYPE`` is used for the ``colType`` parameter and a SQL expression has been supplied
           via the ``expr`` parameter, the method will try to infer the column datatype from the SQL expression when
           the ``build()`` method is called.

           Inferred data types can only be used if the ``expr`` parameter is specified.

           Note that properties which return a schema based on the specification may not be accurate until the
           ``build()`` method is called. Prior to this, the schema may indicate a default column type for those fields.

        You may also add a variety of additional options to further control the test data generation process.
        For full list of options, see :doc:`/reference/api/dbldatagen.column_spec_options`.

        """
        ensure(colName is not None, "Must specify column name for column")
        ensure(colType is not None, f"Must specify column type for column `{colName}`")
        if baseColumn is not None:
            self._checkColumnOrColumnList(baseColumn, allowId=True)

        if not noWarn and colName == self._seedColumnName:
            self.logger.warning(f"Adding a new column named '{colName}' overrides seed column '{self._seedColumnName}'")
            self.logger.warning("Use `seedColumName` option on DataGenerator construction for different seed column")

        # handle migration of old `min` and `max` options
        if _OLD_MIN_OPTION in kwargs:
            assert minValue is None, \
                "Only one of `minValue` and `minValue` can be specified. Use of `minValue` is preferred"
            minValue = kwargs[_OLD_MIN_OPTION]
            kwargs.pop(_OLD_MIN_OPTION, None)

        if _OLD_MAX_OPTION in kwargs:
            assert maxValue is None, \
                "Only one of `maxValue` and `maxValue` can be specified. Use of `maxValue` is preferred"
            maxValue = kwargs[_OLD_MAX_OPTION]
            kwargs.pop(_OLD_MAX_OPTION, None)

        if random is None:
            random = self._defaultRandom

        new_props = {}
        new_props.update(kwargs)

        self.logger.info("effective range: %s, %s, %s args: %s", minValue, maxValue, step, kwargs)
        self.logger.info("adding column - `%s` with baseColumn : `%s`, implicit : %s , omit %s",
                         colName, baseColumn, implicit, omit)
        newColumn = self._generateColumnDefinition(colName, colType, minValue=minValue, maxValue=maxValue,
                                                   step=step, prefix=prefix, random=random,
                                                   distribution=distribution, baseColumn=baseColumn,
                                                   dataRange=dataRange,
                                                   implicit=implicit, omit=omit, **new_props)

        # note for inferred columns, the column type is initially sey to a StringType but may be superceded later
        self._inferredSchemaFields.append(StructField(colName, newColumn.datatype, nullable))
        return self

    def _mkSqlStructFromList(self, fields):
        """
        Create a SQL struct expression from a list of fields

        :param fields: a list of elements that make up the SQL struct expression (each being a string or tuple)
        :returns: SQL expression to generate the struct

        .. note::
          This method is used internally when creating struct columns. It is not intended for general use.

          Each element of the list may be a simple string, or a tuple.
          When the element is specified as a simple string, it must be the name of a previously defined column which
          will be used as both the field name within the struct and the SQL expression to generate the field value.

          When the element is specified as a tuple, it must be a tuple of two elements. The first element must be the
          name of the field within the struct. The second element must be a SQL expression that will be used to generate
          the field value, and may reference previously defined columns.
        """
        assert fields is not None and isinstance(fields, list), \
            "Fields must be a non-empty list of fields that make up the struct elements"
        assert len(fields) >= 1, "Fields must be a non-empty list of fields that make up the struct elements"

        struct_expressions = []

        for fieldSpec in fields:
            if isinstance(fieldSpec, str):
                struct_expressions.append(f"'{fieldSpec}'")
                struct_expressions.append(fieldSpec)
            elif isinstance(fieldSpec, tuple):
                assert len(fieldSpec) == 2, "tuple must be field name and SQL expression strings"
                assert isinstance(fieldSpec[0], str), "First element must be field name string"
                assert isinstance(fieldSpec[1], str), "Second element must be field value SQL string"
                struct_expressions.append(f"'{fieldSpec[0]}'")
                struct_expressions.append(fieldSpec[1])

        struct_expression = f"named_struct({','.join(struct_expressions)})"
        return struct_expression

    def _mkStructFromDict(self, fields):
        assert fields is not None and isinstance(fields, dict), \
            "Fields must be a non-empty dict of fields that make up the struct elements"
        struct_expressions = []

        for key, value in fields.items():
            struct_expressions.append(f"'{key}'")
            if isinstance(value, str):
                struct_expressions.append(str(value))
            elif isinstance(value, dict):
                struct_expressions.append(self._mkStructFromDict(value))
            elif isinstance(value, list):
                array_expressions = ",".join([str(x) for x in value])
                struct_expressions.append(f"array({array_expressions})")
            else:
                raise ValueError(f"Invalid field element for field `{key}`")

        struct_expression = f"named_struct({','.join(struct_expressions)})"
        return struct_expression

    def withStructColumn(self, colName, fields=None, asJson=False, **kwargs):
        """
        Add a struct column to the synthetic data generation specification. This will add a new column composed of
        a struct of the specified fields.

        :param colName: name of column
        :param fields: list of elements to compose as a struct valued column (each being a string or tuple), or a dict
                          outlining the structure of the struct column
        :param asJson: If False, generate a struct valued column. If True, generate a JSON string column
        :param kwargs: keyword arguments to pass to the underlying column generators as per `withColumn`
        :return: A modified in-place instance of data generator allowing for chaining of calls
                  following the Builder pattern

        .. note::
            Additional options for the field specification may be specified as keyword arguments.

            The fields specification specified by the `fields` argument may be :

            - A list of field references (`strings`) which will be used as both the field name and the SQL expression
            - A list of tuples of the form **(field_name, field_expression)** where `field_name` is the name of the
              field. In that case, the `field_expression` string should be a SQL expression to generate the field value
            - A Python dict outlining the structure of the struct column. The keys of the dict are the field names

            When using the `dict` form of the field specifications, a field whose value is a list will be treated
            as creating a SQL array literal.

        """
        assert fields is not None and isinstance(fields, (list, dict)), \
            "Fields argument must be a list of field specifications or dict outlining the target structure "
        assert type(colName) is str and len(colName) > 0, "Must specify a column name"

        if isinstance(fields, list):
            assert len(fields) > 0, \
                "Must specify at least one field for struct column"
            struct_expr = self._mkSqlStructFromList(fields)
        elif isinstance(fields, dict):
            struct_expr = self._mkStructFromDict(fields)
        else:
            raise ValueError(f"Invalid field specification for struct column `{colName}`")

        if asJson:
            output_expr = f"to_json({struct_expr})"
            newDf = self.withColumn(colName, StringType(), expr=output_expr,  **kwargs)
        else:
            newDf = self.withColumn(colName, INFER_DATATYPE, expr=struct_expr, **kwargs)

        return newDf

    def _generateColumnDefinition(self, colName, colType=None, baseColumn=None,
                                  implicit=False, omit=False, nullable=True, **kwargs):
        """ generate field definition and column spec

        .. note:: Any time that a new column definition is added,
                  we'll mark that the build plan needs to be regenerated.
           For our purposes, the build plan determines the order of column generation etc.

        :returns: Newly added column_spec
        """
        if colType is None:
            colType = self.getColumnType(baseColumn)

            if colType == INFER_DATATYPE:
                raise ValueError("When base column(s) have inferred datatype, you must specify the column type")

        new_props = {}
        new_props.update(kwargs)

        # if the column  has the option `random` set to true
        # then use the instance level random seed
        # otherwise use the default random seed for the class
        if OPTION_RANDOM_SEED in new_props:
            effective_random_seed = new_props[OPTION_RANDOM_SEED]
            new_props.pop(OPTION_RANDOM_SEED)
            new_props[OPTION_RANDOM] = True

            # if random seed has override but randomSeedMethod does not
            # set it to fixed
            if OPTION_RANDOM_SEED_METHOD not in new_props:
                new_props[OPTION_RANDOM_SEED_METHOD] = RANDOM_SEED_FIXED

        elif OPTION_RANDOM in new_props and new_props[OPTION_RANDOM]:
            effective_random_seed = self._instanceRandomSeed
        else:
            effective_random_seed = self._randomSeed

        # handle column level override
        if OPTION_RANDOM_SEED_METHOD in new_props:
            effective_random_seed_method = new_props[OPTION_RANDOM_SEED_METHOD]
            new_props.pop(OPTION_RANDOM_SEED_METHOD)
        else:
            effective_random_seed_method = self._seedMethod

        column_spec = ColumnGenerationSpec(colName, colType,
                                           baseColumn=baseColumn,
                                           implicit=implicit,
                                           omit=omit,
                                           randomSeed=effective_random_seed,
                                           randomSeedMethod=effective_random_seed_method,
                                           nullable=nullable,
                                           verbose=self.verbose,
                                           debug=self.debug,
                                           seedColumnName=self._seedColumnName,
                                           **new_props)

        self._columnSpecsByName[colName] = column_spec

        # if column spec for column already exists - remove it
        items_to_remove = [x for x in self._allColumnSpecs if x.name == colName]
        for x in items_to_remove:
            self._allColumnSpecs.remove(x)

        self._allColumnSpecs.append(column_spec)

        # mark that the build plan needs to be regenerated
        self._markForPlanRegen()

        return column_spec

    def _getBaseDataFrame(self, startId=0, streaming=False, options=None):
        """ generate the base data frame and seed column (which defaults to `id`) , partitioning the data if necessary

        This is used when generating the test data.

        A base data frame is created and then each of the additional columns are generated, according to
        base column dependency order, and added to the base data frame using expressions or withColumn statements.

        :returns: Spark data frame for base data that drives the data generation
        """

        end_id = self._rowCount + startId
        id_partitions = self.partitions if self.partitions is not None else 4

        if not streaming:
            status = f"Generating data frame with ids from {startId} to {end_id} with {id_partitions} partitions"
            self.logger.info(status)
            self.executionHistory.append(status)
            df1 = self.sparkSession.range(start=startId,
                                          end=end_id,
                                          numPartitions=id_partitions)

            # spark.range generates a dataframe with the column `id` so rename it if its not our seed column
            if SPARK_RANGE_COLUMN != self._seedColumnName:
                df1 = df1.withColumnRenamed(SPARK_RANGE_COLUMN, self._seedColumnName)

        else:
            status = (
                f"Generating streaming data frame with ids from {startId} to {end_id} with {id_partitions} partitions")
            self.logger.info(status)
            self.executionHistory.append(status)

            df1 = (self.sparkSession.readStream
                   .format("rate"))
            if options is not None:
                if "rowsPerSecond" not in options:
                    options['rowsPerSecond'] = 1
                if "numPartitions" not in options:
                    options['numPartitions'] = id_partitions

                for k, v in options.items():
                    df1 = df1.option(k, v)
                df1 = (df1.load()
                       .withColumnRenamed("value", self._seedColumnName)
                       )

            else:
                df1 = (df1.option("rowsPerSecond", 1)
                       .option("numPartitions", id_partitions)
                       .load()
                       .withColumnRenamed("value", self._seedColumnName)
                       )

        return df1

    def _computeColumnBuildOrder(self):
        """ compute the build ordering using a topological sort on dependencies

        In order to avoid references to columns that have not yet been generated, the test data generation process
        sorts the columns according to the order they need to be built.

        This determines which columns are built first.

        The test generation process will select the columns in the correct order at the end so that the columns
        appear in the correct order in the final output.

        :returns: the build ordering
        """
        dependency_ordering = [(x.name, set(x.dependencies)) if x.name != self._seedColumnName else (
            self._seedColumnName, set())
                               for x in self._allColumnSpecs]

        self.logger.info("dependency list: %s", str(dependency_ordering))

        self._buildOrder = list(
            topologicalSort(dependency_ordering, flatten=False, initial_columns=[self._seedColumnName]))

        self.logger.info("columnBuildOrder: %s", str(self._buildOrder))

        self._buildOrder = self._adjustBuildOrderForSqlDependencies(self._buildOrder, self._columnSpecsByName)

        return self._buildOrder

    def _adjustBuildOrderForSqlDependencies(self, buildOrder, columnSpecsByName):
        """ Adjust column build order according to the following heuristics

        1: if the column being built in a specific build order phase has a SQL expression and it references
           other columns in the same build phase (or potentially references them as the expression parsing is
           primitive), separate that phase into multiple phases.

        It will also issue a warning if the SQL expression appears to reference a column built later

        :param buildOrder: list of lists of ids - each sublist represents phase of build
        :param columnSpecsByName: dictionary to map column names to column specs
        :returns: Spark SQL dataframe of generated test data

        """
        new_build_order = []

        all_columns = set([item for sublist in buildOrder for item in sublist])
        built_columns = []
        prior_phase_built_columns = []

        # for each phase, evaluate it to see if it needs to be split
        for current_phase in buildOrder:
            separate_phase_columns = []

            for columnBeingBuilt in current_phase:

                if columnBeingBuilt in columnSpecsByName:
                    cs = columnSpecsByName[columnBeingBuilt]

                    if cs.expr is not None:
                        sql_references = SchemaParser.columnsReferencesFromSQLString(cs.expr, filterItems=all_columns)

                        # determine references to columns not yet built
                        forward_references = set(sql_references) - set(built_columns)
                        if len(forward_references) > 0:
                            msg = f"Column '{columnBeingBuilt} may have forward references to {forward_references}."
                            self.logger.warning(msg)
                            self.logger.warning("Use `baseColumn` attribute to correct build ordering if necessary")

                        references_not_yet_built = set(sql_references) - set(prior_phase_built_columns)

                        if len(references_not_yet_built.intersection(set(current_phase))) > 0:
                            separate_phase_columns.append(columnBeingBuilt)

                # for each column, get the set of sql references and filter against column names
                built_columns.append(columnBeingBuilt)

            if len(separate_phase_columns) > 0:
                # split phase based on columns in separate_phase_column_list set
                revised_phase = split_list_matching_condition(current_phase,
                                                              lambda el, cols=separate_phase_columns: el in cols)
                new_build_order.extend(revised_phase)
            else:
                # no change to phase
                new_build_order.append(current_phase)

            prior_phase_built_columns.extend(current_phase)

        return new_build_order

    @property
    def build_order(self):
        """ return the build order minus the seed column (which defaults to `id`)

        The build order will be a list of lists - each list specifying columns that can be built at the same time
        """
        if not self.buildPlanComputed:
            self.computeBuildPlan()

        return [x for x in self._buildOrder if x != [self._seedColumnName]]

    def _getColumnDataTypes(self, columns):
        """ Get data types for columns

        :param columns: list of columns to retrieve data types for
        """
        return [self._columnSpecsByName[colspec].datatype for colspec in columns]

    def computeBuildPlan(self):
        """ prepare for building by computing a pseudo build plan

        The build plan is not a true build plan - it is only used for debugging purposes, but does not actually
        drive the column generation order.

        :returns: modified in-place instance of test data generator allowing for chaining of calls
                  following Builder pattern
        """
        self._buildPlan = []
        self.executionHistory = []
        self._processOptions()
        self._buildPlan.append(f"Build Spark data frame with seed column: '{self._seedColumnName}'")

        # add temporary columns
        for cs in self._allColumnSpecs:
            # extend overall build plan with build plan of each column spec
            self._buildPlan.extend(cs._initialBuildPlan)

            # handle generation of any temporary columns
            for tmp_col in cs.temporaryColumns:
                # create column spec for temporary column if its not already present
                if not self.hasColumnSpec(tmp_col[0]):
                    self._buildPlan.append(f"materializing temporary column {tmp_col[0]}")
                    self.withColumn(tmp_col[0], tmp_col[1], **tmp_col[2])

        # TODO: set up the base column data type information
        for cs in self._allColumnSpecs:
            base_column_datatypes = self._getColumnDataTypes(cs.baseColumns)
            cs.setBaseColumnDatatypes(base_column_datatypes)

        self._computeColumnBuildOrder()

        for x1 in self._buildOrder:
            for x in x1:
                cs = self._columnSpecsByName[x]
                self._buildPlan.append(cs.getPlanEntry())

        self.buildPlanComputed = True
        return self

    def build(self, withTempView=False, withView=False, withStreaming=False, options=None):
        """ build the test data set from the column definitions and return a dataframe for it

        if `withStreaming` is True, generates a streaming data set.
        Use options to control the rate of generation of test data if streaming is used.

        For example:

        `dfTestData = testDataSpec.build(withStreaming=True,options={ 'rowsPerSecond': 5000})`

        :param withTempView: if True, automatically creates temporary view for generated data set
        :param withView: If True, automatically creates global view for data set
        :param withStreaming: If True, generates data using Spark Structured Streaming Rate source suitable
                              for writing with `writeStream`
        :param options: optional Dict of options to control generating of streaming data
        :returns: Spark SQL dataframe of generated test data
        """
        self.logger.debug("starting build ... withStreaming [%s]", withStreaming)
        self.executionHistory = []
        self.computeBuildPlan()

        output_columns = self.getOutputColumnNames()
        ensure(output_columns is not None and len(output_columns) > 0,
               """
                | You must specify at least one column for output
                | - use withIdOutput() to output base seed column
               """)

        df1 = self._getBaseDataFrame(self.starting_id, streaming=withStreaming, options=options)

        self.executionHistory.append("Using Pandas Optimizations {True}")

        # build columns
        df1 = self._buildColumnExpressionsWithSelects(df1)

        df1 = df1.select(*self.getOutputColumnNames())
        self.executionHistory.append(f"selecting columns: {self.getOutputColumnNames()}")

        # register temporary or global views if necessary
        if withView:
            self.executionHistory.append("registering view")
            self.logger.info("Registered global view [%s]", self.name)
            df1.createGlobalTempView(self.name)
            self.logger.info("Registered!")
        elif withTempView:
            self.executionHistory.append("registering temp view")
            self.logger.info("Registering temporary view [%s]", self.name)
            df1.createOrReplaceTempView(self.name)
            self.logger.info("Registered!")

        return df1

    # noinspection PyProtectedMember
    def _buildColumnExpressionsWithSelects(self, df1):
        """
        Build column generation expressions with selects
        :param df1: dataframe for base data generator
        :return: new dataframe

        The data generator build plan is separated into `rounds` of expressions. Each round consists of
        expressions that are generated using a single `select` operation
        """
        self.executionHistory.append("Generating data with selects")
        # generation with selects may be more efficient as less intermediate data frames
        # are generated resulting in shorter lineage
        for colNames in self.build_order:
            build_round = ["*"]
            column_specs_applied = []
            inx_col = 0
            self.executionHistory.append(f"building stage for columns: {colNames}")
            for colName in colNames:
                col1 = self._columnSpecsByName[colName]
                column_generators = col1.makeGenerationExpressions()
                self.executionHistory.extend(col1.executionHistory)
                if type(column_generators) is list and len(column_generators) == 1:
                    build_round.append(column_generators[0].alias(colName))
                elif type(column_generators) is list and len(column_generators) > 1:
                    i = 0
                    for cg in column_generators:
                        build_round.append(cg.alias(f'{colName}_{i}'))
                        i += 1
                else:
                    build_round.append(column_generators.alias(colName))
                column_specs_applied.append(col1)
                inx_col = inx_col + 1

            df1 = df1.select(*build_round)

            # apply any post select processing
            for cs in column_specs_applied:
                cs._onSelect(df1)
        return df1

    def _sqlTypeFromSparkType(self, dt):
        """Get sql type for spark type
           :param dt: instance of Spark SQL type such as IntegerType()
        """
        return dt.simpleString()

    def _mkInsertOrUpdateStatement(self, columns, srcAlias, substitutions, isUpdate=True):
        if substitutions is None:
            substitutions = []
        results = []
        subs = {}
        for x in columns:
            subs[x] = f"{srcAlias}.{x}"
        for x in substitutions:
            subs[x[0]] = x[1]

        for substitution_col in columns:
            new_val = subs[substitution_col]
            if isUpdate:
                results.append(f"{substitution_col}={new_val}")
            else:
                results.append(f"{new_val}")

        return ", ".join(results)

    def scriptTable(self, name=None, location=None, tableFormat="delta", asHtml=False):
        """ generate create table script suitable for format of test data set

        :param name: name of table to use in generated script
        :param location: path to location of data. If specified (default is None), will generate
                         an external table definition.
        :param tableFormat: table format for table
        :param asHtml: if true, generate output suitable for use with `displayHTML` method in notebook environment
        :returns: SQL string for scripted table
        """
        assert name is not None, "`name` must be specified"

        self.computeBuildPlan()

        output_columns = self.getOutputColumnNamesAndTypes()

        results = [f"CREATE TABLE IF NOT EXISTS {name} ("]
        ensure(output_columns is not None and len(output_columns) > 0,
               """
                | You must specify at least one column for output
                | - use withIdOutput() to output base seed column
               """)

        col_expressions = []
        for col_to_output in output_columns:
            col_expressions.append(f"    {col_to_output[0]} {self._sqlTypeFromSparkType(col_to_output[1])}")
        results.append(",\n".join(col_expressions))
        results.append(")")
        results.append(f"using {tableFormat}")

        if location is not None:
            results.append(f"location '{location}'")

        results = "\n".join(results)

        if asHtml:
            results = HtmlUtils.formatCodeAsHtml(results)

        return results

    def scriptMerge(self, tgtName=None, srcName=None, updateExpr=None, delExpr=None, joinExpr=None, timeExpr=None,
                    insertExpr=None,
                    useExplicitNames=True,
                    updateColumns=None, updateColumnExprs=None,
                    insertColumns=None, insertColumnExprs=None,
                    srcAlias="src", tgtAlias="tgt",
                    asHtml=False
                    ):
        """ generate merge table script suitable for format of test data set

        :param tgtName: name of target table to use in generated script
        :param tgtAlias: alias for target table - defaults to `tgt`
        :param srcName: name of source table to use in generated script
        :param srcAlias: alias for source table - defaults to `src`
        :param updateExpr: optional string representing updated condition. If not present, then
                            any row that does not match join condition is considered an update
        :param delExpr: optional string representing delete condition - For example `src.action='DEL'`.
                         If not present, no delete clause is generated
        :param insertExpr: optional string representing insert condition - If not present,
                        there is no condition on insert other than no match
        :param joinExpr: string representing join condition. For example, `tgt.id=src.id`
        :param timeExpr: optional time travel expression - for example : `TIMESTAMP AS OF timestamp_expression`
                        or `VERSION AS OF version`
        :param insertColumns: Optional list of strings designating columns to insert.
                               If not supplied, uses all columns defined in spec
        :param insertColumnExprs: Optional list of strings designating designating column expressions for insert.
            By default, will use src column as insert value into
            target table. This should have the form [ ("insert_column_name", "insert column expr"), ...]
        :param updateColumns: List of strings designating columns to update.
                               If not supplied, uses all columns defined in spec
        :param updateColumnExprs: Optional list of strings designating designating column expressions for update.
            By default, will use src column as update value for
            target table. This should have the form [ ("update_column_name", "update column expr"), ...]
        :param useExplicitNames: If True, generate explicit column names in insert and update statements
        :param asHtml: if true, generate output suitable for use with `displayHTML` method in notebook environment
        :returns: SQL string for scripted merge statement
        """
        assert tgtName is not None, "you must specify a target table"
        assert srcName is not None, "you must specify a source table"
        assert joinExpr is not None, "you must specify a join expression"

        self.computeBuildPlan()

        # get list of column names
        output_columns = [x[0] for x in self.getOutputColumnNamesAndTypes()]

        ensure(output_columns is not None and len(output_columns) > 0,
               """
                | You must specify at least one column for output
                | - use withIdOutput() to output base seed column
               """)

        # use list of column names if not supplied
        if insertColumns is None:
            insertColumns = output_columns

        if updateColumns is None:
            updateColumns = output_columns

        # build merge statement
        results = [f"MERGE INTO `{tgtName}` as {tgtAlias}"]

        # use time expression if supplied
        if timeExpr is None:
            results.append(f"USING `{srcName}` as {srcAlias}")
        else:
            results.append(f"USING `{srcName}` {timeExpr} as {srcAlias}")

        # add join condition
        results.append(f"ON {joinExpr}")

        # generate update clause
        update_clause = None
        if updateExpr is not None:
            update_clause = f"WHEN MATCHED and {updateExpr} THEN UPDATE "
        else:
            update_clause = "WHEN MATCHED THEN UPDATE "

        if not useExplicitNames:
            update_clause = update_clause + " SET *"
        else:
            update_clause = (update_clause
                             + " SET "
                             + self._mkInsertOrUpdateStatement(columns=updateColumns, srcAlias=srcAlias,
                                                               substitutions=updateColumnExprs))

        results.append(update_clause)

        # generate delete clause
        if delExpr is not None:
            results.append(f"WHEN MATCHED and {delExpr} THEN DELETE")

        if insertExpr is not None:
            ins_clause = f"WHEN NOT MATCHED and {insertExpr} THEN INSERT "
        else:
            ins_clause = "WHEN NOT MATCHED THEN INSERT "

        if not useExplicitNames:
            ins_clause = ins_clause + " *"
        else:
            ins_clause = (ins_clause + "(" + ",".join(insertColumns)
                          + ") VALUES (" +
                          self._mkInsertOrUpdateStatement(columns=insertColumns, srcAlias=srcAlias,
                                                          substitutions=insertColumnExprs, isUpdate=False)
                          + ")"
                          )

        results.append(ins_clause)

        result = "\n".join(results)

        if asHtml:
            result = HtmlUtils.formatCodeAsHtml(results)

        return result
